#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Fetcher.py, module definition of FoBiS.py Fetcher object.

This module handles fetching and building GitHub-hosted FoBiS.py dependencies.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
try:
  import configparser as configparser
except ImportError:
  import configparser
import os
import shutil
from .utils import print_fake, syswork


class Fetcher(object):
  """
  Fetcher handles fetching and building GitHub-hosted FoBiS.py dependencies.
  """
  DEPS_CONFIG_FILE = '.deps_config.ini'

  def __init__(self, deps_dir, print_n=None, print_w=None):
    """
    Parameters
    ----------
    deps_dir : str
      directory where dependencies are stored
    print_n : {None}
      function for printing normal messages
    print_w : {None}
      function for printing warning messages
    """
    self.deps_dir = deps_dir
    self.print_n = print_n if print_n is not None else print_fake
    self.print_w = print_w if print_w is not None else print_fake

  def parse_dep_spec(self, spec):
    """
    Parse a dependency spec string into a dict.

    Format: URL [:: branch=X] [:: tag=X] [:: rev=X] [:: mode=X]

    Parameters
    ----------
    spec : str
      dependency specification string

    Returns
    -------
    dict
      parsed fields: 'url', and optionally 'branch', 'tag', 'rev', 'mode'
    """
    parts = [p.strip() for p in spec.split('::')]
    result = {'url': parts[0].strip()}
    for part in parts[1:]:
      if '=' in part:
        key, _, value = part.partition('=')
        result[key.strip()] = value.strip()
    return result

  def fetch(self, name, url, branch=None, tag=None, rev=None, update=False):
    """
    Clone or update a dependency repo.

    Parameters
    ----------
    name : str
      dependency name (used as subdirectory name)
    url : str
      git repository URL
    branch : {None}
      branch to check out
    tag : {None}
      tag to check out
    rev : {None}
      revision/commit to check out
    update : {False}
      if True, run git fetch to update before checking out

    Returns
    -------
    str
      path to the cloned dependency directory
    """
    dep_dir = os.path.join(self.deps_dir, name)
    if not os.path.exists(self.deps_dir):
      os.makedirs(self.deps_dir)
    if not os.path.exists(dep_dir):
      self.print_n('Cloning ' + name + ' from ' + url)
      result = syswork('git clone ' + url + ' ' + dep_dir)
      if result[0] != 0:
        self.print_w('Error cloning ' + name + ':\n' + result[1])
        return dep_dir
    elif update:
      self.print_n('Updating ' + name)
      result = syswork('git -C ' + dep_dir + ' fetch')
      if result[0] != 0:
        self.print_w('Error fetching updates for ' + name + ':\n' + result[1])
    else:
      self.print_n('Dependency ' + name + ' already fetched (use --update to refresh)')
    ref = tag or branch or rev
    if ref:
      self.print_n('Checking out ' + ref + ' for ' + name)
      result = syswork('git -C ' + dep_dir + ' checkout ' + ref)
      if result[0] != 0:
        self.print_w('Error checking out ' + ref + ' for ' + name + ':\n' + result[1])
    elif update:
      result = syswork('git -C ' + dep_dir + ' merge --ff-only')
      if result[0] != 0:
        self.print_w('Error merging updates for ' + name + ':\n' + result[1])
    return dep_dir

  def build_dep(self, name, dep_dir, mode=None):
    """
    Build a dependency using FoBiS.py build.

    Parameters
    ----------
    name : str
      dependency name (for messages)
    dep_dir : str
      path to the dependency directory
    mode : {None}
      fobos mode to use for building
    """
    fobos_file = os.path.join(dep_dir, 'fobos')
    if not os.path.exists(fobos_file):
      self.print_w('Error: dependency "' + name + '" has no fobos file in ' + dep_dir)
      return
    self.print_n('Building dependency ' + name)
    old_pwd = os.getcwd()
    os.chdir(dep_dir)
    if mode:
      result = syswork('FoBiS.py build -mode ' + mode)
    else:
      result = syswork('FoBiS.py build')
    os.chdir(old_pwd)
    if result[0] != 0:
      self.print_w('Error building ' + name + ':\n' + result[1])
    else:
      self.print_n(result[1])

  def save_config(self, deps_info):
    """
    Write the deps config file for use by FoBiS.py build.

    Deps with use='fobos' are written as 'dependon' entries (library approach).
    Deps with use='sources' (default) are written as 'src' entries (direct source inclusion).

    Parameters
    ----------
    deps_info : list
      list of dicts with keys 'name', 'path', 'mode', 'use'
    """
    config = configparser.RawConfigParser()
    config.add_section('deps')
    dependon_entries = []
    src_entries = []
    for dep in deps_info:
      if dep.get('use', 'sources') == 'fobos':
        fobos_path = os.path.join(dep['path'], 'fobos')
        entry = fobos_path + ':' + dep['mode'] if dep.get('mode') else fobos_path
        dependon_entries.append(entry)
      else:
        src_entries.append(dep['path'])
    if dependon_entries:
      config.set('deps', 'dependon', ' '.join(dependon_entries))
    if src_entries:
      config.set('deps', 'src', ' '.join(src_entries))
    config_path = os.path.join(self.deps_dir, self.DEPS_CONFIG_FILE)
    with open(config_path, 'w') as cfg_file:
      config.write(cfg_file)
    self.print_n('Saved deps config to ' + config_path)

  def _resolve_url(self, repo):
    """
    Convert a shorthand repo reference to a full GitHub HTTPS URL.

    Accepts 'user/repo' (GitHub shorthand) or any full URL unchanged.

    Parameters
    ----------
    repo : str
      repository reference

    Returns
    -------
    str
      full git-cloneable URL
    """
    if repo.startswith('http://') or repo.startswith('https://') or repo.startswith('git@'):
      return repo
    return 'https://github.com/' + repo

  def install_from_github(self, repo, branch=None, tag=None, rev=None, mode=None,
                          update=False, no_build=False,
                          prefix='./', bin_dir='bin/', lib_dir='lib/', include_dir='include/'):
    """
    Clone, build, and install a GitHub-hosted FoBiS project.

    Parameters
    ----------
    repo : str
      repository reference ('user/repo' shorthand or full URL)
    branch : {None}
      branch to check out
    tag : {None}
      tag to check out
    rev : {None}
      revision/commit to check out
    mode : {None}
      fobos mode to use when building
    update : {False}
      re-fetch before building
    no_build : {False}
      clone only, skip building and installing
    prefix : str
      installation prefix directory
    bin_dir : str
      sub-directory under prefix for executables
    lib_dir : str
      sub-directory under prefix for libraries
    include_dir : str
      sub-directory under prefix for module files
    """
    url = self._resolve_url(repo)
    name = url.rstrip('/').split('/')[-1]
    if name.endswith('.git'):
      name = name[:-4]
    dep_dir = self.fetch(name, url, branch=branch, tag=tag, rev=rev, update=update)
    if no_build:
      self.print_n('Cloned ' + name + ' to ' + dep_dir + ' (--no-build: skipping build and install)')
      return
    self._build_dep_tracked(name, dep_dir, mode=mode)
    self._install_artifacts(dep_dir, prefix, bin_dir, lib_dir, include_dir)

  def _build_dep_tracked(self, name, dep_dir, mode=None):
    """
    Build a dependency using 'fobis build --track_build'.

    Parameters
    ----------
    name : str
      dependency name (for messages)
    dep_dir : str
      path to the dependency directory
    mode : {None}
      fobos mode to use for building
    """
    fobos_file = os.path.join(dep_dir, 'fobos')
    if not os.path.exists(fobos_file):
      self.print_w('Error: "' + name + '" has no fobos file in ' + dep_dir)
      return
    self.print_n('Building ' + name + ' (with --track_build)')
    old_pwd = os.getcwd()
    os.chdir(dep_dir)
    cmd = 'fobis build --track_build'
    if mode:
      cmd += ' -mode ' + mode
    result = syswork(cmd)
    os.chdir(old_pwd)
    if result[0] != 0:
      self.print_w('Error building ' + name + ':\n' + result[1])
    else:
      self.print_n(result[1])

  def _install_artifacts(self, dep_dir, prefix, bin_dir, lib_dir, include_dir):
    """
    Scan dep_dir for .track_build files and copy artifacts to the prefix.

    Parameters
    ----------
    dep_dir : str
      root of the cloned dependency
    prefix : str
      installation prefix
    bin_dir : str
      sub-directory under prefix for executables
    lib_dir : str
      sub-directory under prefix for libraries
    include_dir : str
      sub-directory under prefix for module/include files
    """
    installed_any = False
    for root, _, files in os.walk(dep_dir):
      for filename in files:
        if not filename.endswith('.track_build'):
          continue
        track_file = configparser.ConfigParser()
        track_file.read(os.path.join(root, filename))
        if not track_file.has_option('build', 'output'):
          continue
        output = track_file.get('build', 'output')
        if not os.path.exists(output):
          continue
        is_program = (track_file.has_option('build', 'program') and
                      track_file.get('build', 'program'))
        is_library = (track_file.has_option('build', 'library') and
                      track_file.get('build', 'library'))
        if is_program:
          dest = os.path.join(prefix, bin_dir)
          os.makedirs(dest, exist_ok=True)
          self.print_n('Installing "' + output + '" -> "' + dest + '"')
          shutil.copy(output, dest)
          installed_any = True
        if is_library:
          dest = os.path.join(prefix, lib_dir)
          os.makedirs(dest, exist_ok=True)
          self.print_n('Installing "' + output + '" -> "' + dest + '"')
          shutil.copy(output, dest)
          installed_any = True
          if track_file.has_option('build', 'mod_file'):
            mod_file = track_file.get('build', 'mod_file')
            inc_dest = os.path.join(prefix, include_dir)
            os.makedirs(inc_dest, exist_ok=True)
            self.print_n('Installing "' + mod_file + '" -> "' + inc_dest + '"')
            shutil.copy(mod_file, inc_dest)
    if not installed_any:
      self.print_w('No installable artifacts found. Ensure the project fobos uses --track_build or check the fobos mode.')

  def load_config(self):
    """
    Read the deps config file and return a dict with 'dependon' and 'src' lists.

    Returns
    -------
    dict
      'dependon' : list of fobos-path strings for use=fobos deps
      'src'      : list of directory paths for use=sources deps
      either key is absent when no entries of that type exist
    """
    config_path = os.path.join(self.deps_dir, self.DEPS_CONFIG_FILE)
    if not os.path.exists(config_path):
      return {}
    config = configparser.RawConfigParser()
    config.read(config_path)
    result = {}
    if config.has_option('deps', 'dependon'):
      result['dependon'] = config.get('deps', 'dependon').split()
    if config.has_option('deps', 'src'):
      result['src'] = config.get('deps', 'src').split()
    return result



"""
Gcov.py, module definition of Gcov class.

This is a class designed for analyzing gcov files.
"""
# from __future__ import division
# from __future__ import absolute_import
# from __future__ import print_function
# from __future__ import unicode_literals
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
# from builtins import object
try:
  from pygooglechart import PieChart2D
  __graph__ = True
except ImportError:
  # print("Module 'pygooglechart' not found: Gcov graphs disabled")
  __graph__ = False
import os
import re
import sys
from operator import itemgetter
__str_kw_elemental__ = r"[Ee][Ll][Ee][Mm][Ee][Nn][Tt][Aa][Ll]"
__str_kw_pure__ = r"[Pp][Uu][Rr][Ee]"
__str_kw_subroutine__ = r"[Ss][Uu][Bb][Rr][Oo][Uu][Tt][Ii][Nn][Ee]"
__str_kw_function__ = r"[Ff][Uu][Nn][Cc][Tt][Ii][Oo][Nn]"
__str_name__ = r"(?P<name>[a-zA-Z][a-zA-Z0-9_]*)"
__str_procedure__ = (r"^(\s*)" +  # initial white spaces
                     r"(" + __str_kw_elemental__ + r"|" + __str_kw_pure__ + r")?\s+" +  # "elemental" or "pure" attribute
                     r"(?P<ptype>" + __str_kw_function__ + r"|" + __str_kw_subroutine__ + r")\s+" +  # type of procedure
                     r"(?P<pname>" + __str_name__ + r"\s*).*")  # name of procedure
__regex_procedure__ = re.compile(__str_procedure__)


def _mermaid_pie(title, executed, unexecuted):
  """Return a Mermaid pie-chart fenced code block string with surrounding blank lines."""
  return ('```mermaid\npie showData\n' +
          '    title ' + title + '\n' +
          '    "Executed" : ' + executed + '\n' +
          '    "Unexecuted" : ' + unexecuted + '\n' +
          '```\n\n')


class Gcov(object):
  """Gcov is an object that handles gcov file, its attributes and methods."""

  def __init__(self, filename=None):
    """
    Parameters
    ----------
    filename : {None}
      string containing the gcov filename
    """
    self.filename = filename
    self.procedures = []
    self.coverage = []
    self.metrics = {'coverage': None, 'procedures': None}
    self.l_pie_url = ''
    self.p_pie_url = ''
    return

  def _metrics(self):
    """
    Getg gcov metrics.
    """
    if len(self.coverage) > 0:
      lnumber = sum(1 for cov in self.coverage if isinstance(cov, int))
      elnumber = sum(cov > 0 for cov in self.coverage if isinstance(cov, int))
      unelnumber = sum(cov == 0 for cov in self.coverage if isinstance(cov, int))
      if lnumber > 0:
        elnumber_per = int(round(elnumber * 100. / lnumber))
        unelnumber_per = int(round(unelnumber * 100. / lnumber))
      else:
        elnumber_per = 0
        unelnumber_per = 0
      if elnumber > 0:
        ahits = sum(cov for cov in self.coverage if isinstance(cov, int)) / elnumber
      else:
        ahits = 0
      self.metrics['coverage'] = [str(lnumber), str(elnumber), str(unelnumber), str(elnumber_per), str(unelnumber_per), str(ahits)]
    if len(self.procedures) > 0:
      pnumber = len(self.procedures)
      epnumber = sum(proc[2] > 0 for proc in self.procedures)
      unepnumber = sum(proc[2] == 0 for proc in self.procedures)
      if pnumber > 0:
        epnumber_per = int(round(epnumber * 100. / pnumber))
        unepnumber_per = int(round(unepnumber * 100. / pnumber))
      else:
        epnumber_per = 0
        unepnumber_per = 0
      if epnumber > 0:
        ahits = sum(proc[2] for proc in self.procedures) / epnumber
      else:
        ahits = 0
      self.metrics['procedures'] = [str(pnumber), str(epnumber), str(unepnumber), str(epnumber_per), str(unepnumber_per), str(ahits)]
    return

  def _get_graphs_url(self):
    """
    Get pie-chart graphs urls from online google chart service.

    Returns
    ----------
    str, str
      string containing the url of online google pie-chart graphs
    """
    def pie(size, title, labels, legend, data, colors):
      """
      Function for creating a pie chart.

      Parameters
      ----------
      size : list
        width and height
      title : str
        title of the graph
      lables : list
        list of string containing labels of the data
      legend : list
        list of string containing legend of the data
      data : list
        data to be plotted
      colors : list
        list of string containing the colors of each data in PRGB format

      Returns
      -------
      PieChart2D object
      """
      graph = PieChart2D(size[0], size[1])
      graph.set_title(title)
      graph.set_legend(legend)
      graph.set_pie_labels(labels)
      graph.add_data(data)
      graph.set_colours(colors)
      return graph

    if __graph__:
      if self.metrics['coverage']:
        l_pie = pie(size=[300, 150],
                    title='Coverage of executable lines',
                    labels=[self.metrics['coverage'][3] + '%', self.metrics['coverage'][4] + '%'],
                    legend=['Executed', 'Unexecuted'],
                    data=[int(self.metrics['coverage'][3]), int(self.metrics['coverage'][4])],
                    colors=['65C1FF', 'FF9260'])
        self.l_pie_url = '[![lines](' + l_pie.get_url() + ')]()'
      if self.metrics['procedures']:
        p_pie = pie(size=[300, 150],
                    title='Coverage of procedures',
                    labels=[self.metrics['procedures'][3] + '%', self.metrics['procedures'][4] + '%'],
                    legend=['Executed', 'Unexecuted'],
                    data=[int(self.metrics['procedures'][3]), int(self.metrics['procedures'][4])],
                    colors=['65C1FF', 'FF9260'])
        self.p_pie_url = '[![procedures](' + p_pie.get_url() + ')]()'
    return

  def parse(self):
    """
    Parse gcov file.
    """
    def _coverage_update(ignoring, cov_num, line_num, text, coverage):
      """
      Update coverage list.
      """
      if re.search(r'\bLCOV_EXCL_START\b', text):
        if ignoring:
          sys.stderr.write("Warning: %s:%d: nested LCOV_EXCL_START, "
                           "please fix\n" % (self.filename, line_num))
        ignoring = True
      elif re.search(r'\bLCOV_EXCL_END\b', text):
        if not ignoring:
          sys.stderr.write("Warning: %s:%d: LCOV_EXCL_END outside of "
                           "exclusion zone, please fix\n" % (self.filename,
                                                             line_num))
        ignoring = False
      if cov_num == '-':
        coverage.append(None)
      elif cov_num == '#####':
        if (
            ignoring or
            text.lstrip().startswith(('inline', 'static')) or
            text.strip() == '}' or
            re.search(r'\bLCOV_EXCL_LINE\b', text)
        ):
          coverage.append(None)
        else:
          coverage.append(0)
      elif cov_num == '=====':
        coverage.append(0)
      elif '*' in cov_num:
        coverage.append(int(cov_num.rstrip('*')))
      else:
        try:
          cov_num_int = int(cov_num) 
        except ValueError:
          cov_num_int = 0
        coverage.append(cov_num_int)
      return ignoring

    ignoring = False
    if self.filename:
      if os.path.exists(self.filename):
        with open(self.filename, 'r') as source:
          for line in source:
            if ":" not in line:
              break
            report_fields = line.split(':')
            cov_num = report_fields[0].strip()
            line_num = int(report_fields[1].strip())
            text = report_fields[2]
            proc_matching = re.match(__regex_procedure__, text)
            if line_num == 0:
              continue
            if proc_matching:
              ptype = proc_matching.group('ptype').strip()
              pname = proc_matching.group('pname').strip()
              if cov_num != '#####' and cov_num != '-':
                try:
                  pcov = int(cov_num.strip('*'))
                except ValueError:
                  pcov = 0
              else:
                pcov = 0
              if cov_num != '-':  # needed due to abstract iterfaces
                self.procedures.append([ptype, pname, pcov, line_num])
            ignoring = _coverage_update(ignoring=ignoring, cov_num=cov_num, line_num=line_num, text=text, coverage=self.coverage)
    self._metrics()
    self._get_graphs_url()
    return

  def save(self, report_format='markdown', output=None, graphs=False):
    """
    Save gcov report analysis.

    Parameters
    ----------
    report_format : {'markdown'}
      output format of saved report
    output : {None}
      string output file name
    graphs : {False}
      flag for including graphs
    """
    string = []
    if len(self.metrics) > 0:
      if report_format == 'markdown':
        string.append('### Coverage analysis of *' + os.path.splitext(os.path.basename(self.filename))[0] + '*\n')
        string.append('\n')
        if self.metrics['coverage']:
          string.append('|Lines| | |\n')
          string.append('| --- | --- | --- |\n')
          string.append('|Executable lines            |' + self.metrics['coverage'][0] + '| |\n')
          string.append('|Executed lines              |' + self.metrics['coverage'][1] + '|' + self.metrics['coverage'][3] + '%|\n')
          string.append('|Unexecuted lines            |' + self.metrics['coverage'][2] + '|' + self.metrics['coverage'][4] + '%|\n')
          string.append('|Average hits / executed     |' + self.metrics['coverage'][5] + '| |\n')
          string.append('\n')
          string.append(_mermaid_pie('Lines (' + self.metrics['coverage'][3] + '% covered)',
                                     self.metrics['coverage'][1], self.metrics['coverage'][2]))
        if self.metrics['procedures']:
          string.append('|Procedures| | |\n')
          string.append('| --- | --- | --- |\n')
          string.append('|Total procedures            |' + self.metrics['procedures'][0] + '| |\n')
          string.append('|Executed procedures         |' + self.metrics['procedures'][1] + '|' + self.metrics['procedures'][3] + '%|\n')
          string.append('|Unexecuted procedures       |' + self.metrics['procedures'][2] + '|' + self.metrics['procedures'][4] + '%|\n')
          string.append('|Average hits / executed     |' + self.metrics['procedures'][5] + '| |\n')
          string.append('\n')
          string.append(_mermaid_pie('Procedures (' + self.metrics['procedures'][3] + '% covered)',
                                     self.metrics['procedures'][1], self.metrics['procedures'][2]))
        if graphs and __graph__ and (self.metrics['coverage'] or self.metrics['procedures']):
          string.append('\n --- \n')
          if self.metrics['coverage']:
            string.append(self.l_pie_url + '\n')
          if self.metrics['procedures']:
            string.append(self.p_pie_url + '\n')
        string.append('\n#### Unexecuted procedures\n')
        string.append('\n')
        if any(proc[2] == 0 for proc in self.procedures):
          for proc in sorted(self.procedures, key=itemgetter(0, 1)):
            if proc[2] == 0:
              string.append(' + *' + proc[0] + '* **' + proc[1] + '**, line ' + str(proc[3]) + '\n')
        else:
          string.append(' + *none*\n')
        string.append('\n#### Executed procedures\n')
        string.append('\n')
        if any(proc[2] > 0 for proc in self.procedures):
          for proc in sorted(self.procedures, key=itemgetter(2), reverse=True):
            if proc[2] > 0:
              string.append(' + *' + proc[0] + '* **' + proc[1] + '**: tested **' + str(proc[2]) + '** times\n')
        else:
          string.append(' + *none*\n')
        string.append('\n --- \n Report generated by [FoBiS.py](https://github.com/szaghi/FoBiS)')
    if output:
      fout = output
    else:
      fout = os.path.splitext(os.path.basename(self.filename))[0] + '.gcov-report.md'
    with open(fout, 'w') as out:
      out.writelines(string)
    return



"""
utils.py, module definition of FoBiS.py util functions.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from __future__ import print_function
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import unicode_literals
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
import os
import subprocess
import sys


def print_fake(input_obj, end='\n'):
  """Rename print."""
  print(input_obj, end=end)


def syswork(cmd):
  """
  Function for executing system command 'cmd': for compiling and linking files.
  """
  error = 0
  try:
    output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
  except subprocess.CalledProcessError as err:
    error = err.returncode
    output = err.output
  if sys.version_info[0] > 2:
    output = str(output, encoding='UTF-8')
  return [error, str(output)]


def traverse_recursive(parsed_file, path=None):
  """
  The function traverse_recursive performs a yield-recursive traversing of pfile direct dependencies.
  """
  if path is None:
    path = list()
  path.append(parsed_file)
  yield path
  for dep in parsed_file.pfile_dep:
    if dep != parsed_file:
      for path in traverse_recursive(dep, path):
        yield path
  if path:
    path.pop()


def unique_seq(seq):
  """
  The function unique_seq returns the input sequence removing duplicated elements but peserving the original order.
  """
  seen = set()
  seen_add = seen.add
  return [x for x in seq if x not in seen and not seen_add(x)]


def module_is_in(pfiles, module):
  """
  Find the parsed file containing the desidered module.

  Parameters
  ----------
  pfiles : list
    list of parsed files
  module : str
    module name

  Returns
  -------
  file_name : str
    name of file containing the module
  number : int
    number of file containing the module
  """
  file_name = ""
  number = -1
  for fnum, parsed_file in enumerate(pfiles):
    if parsed_file.module:
      for module_name in parsed_file.module_names:
        if module_name.lower() == module.lower():
          file_name = parsed_file.name
          number = fnum
          break
  return file_name, number


def include_is_in(pfiles, include):
  """
  Function finding the parsed file containing the desidered include-file.

  Parameters
  ----------
  pfiles : list
    list of parsed files
  include : str
    include-file name

  Returns
  -------
  file_name : str
    name of file containing the include-file
  number : int
    number of file containing the include-file
  """
  file_name = ""
  number = -1
  for fnum, parsed_file in enumerate(pfiles):
    if os.path.basename(parsed_file.name) == include:
      file_name = parsed_file.name
      number = fnum
      break
  return file_name, number


def dependency_hiearchy(builder, pfiles, print_w=None, force_compile=False):
  """
  Build parsed files hierarchy.

  Paramters
  ---------
  builder : Builder object
  pfiles : list
    list of ParsedFile objects
  print_w : function
    function for printing emphized warning message
  force_compile : {False}
    flag for forcing (re-)compiling of all dependency
  """
  if print_w is None:
    print_wm = print_fake
  else:
    print_wm = print_w

  # direct dependencies list used after for building indirect (complete) dependencies list
  for parsed_file in pfiles:
    parsed_file.pfile_dep = []
    for dep in parsed_file.dependencies:
      if dep.type == "module":
        dep.file, fnum = module_is_in(pfiles=pfiles, module=dep.name)
        if fnum > -1:
          if not pfiles[fnum] in parsed_file.pfile_dep:
            parsed_file.pfile_dep.append(pfiles[fnum])
        else:
          # there is no source containg the searched module, try to find it into a precompiled mod file into eventually included paths
          for dinc in builder.dinc:
            for _, _, files in os.walk(dinc):
              for filename in files:
                if os.path.splitext(os.path.basename(filename))[0].lower() == dep.name.lower() and os.path.splitext(os.path.basename(filename))[1].lower() == '.mod':
                  fnum = 0
                  break
              if fnum == 0:
                break
            if fnum == 0:
              break
          if fnum != 0:
            print_wm("Attention: the file '" + parsed_file.name + "' depends on '" + dep.name + "' that is unreachable")
            # sys.exit(1)
      if dep.type == "include":
        dep.file, fnum = include_is_in(pfiles=pfiles, include=dep.name)
        if fnum > -1:
          if not pfiles[fnum] in parsed_file.pfile_dep:
            pfiles[fnum].program = False
            pfiles[fnum].module = False
            pfiles[fnum].nomodlib = False
            pfiles[fnum].include = True
            parsed_file.pfile_dep.append(pfiles[fnum])
            if not os.path.dirname(pfiles[fnum].name) in builder.dinc:
              builder.dinc.append(os.path.dirname(pfiles[fnum].name))
        else:
          print_wm("Attention: the file '" + parsed_file.name + "' depends on '" + dep.name + "' that is unreachable")
          # sys.exit(1)
  # indirect dependency list
  for parsed_file in pfiles:
    parsed_file.create_pfile_dep_all()
  # using the just created hiearchy for checking which files must be (re-)compiled
  for parsed_file in pfiles:
    parsed_file.check_compile(obj_dir=builder.obj_dir, force_compile=force_compile)


def remove_other_main(builder, pfiles, myself):
  """
  The function remove_other_main removes all compiled objects of other program than the current target under building.
  """
  for parsed_file in pfiles:
    if parsed_file.program and parsed_file.name != myself.name:
      if os.path.exists(builder.obj_dir + parsed_file.basename + ".o"):
        os.remove(builder.obj_dir + parsed_file.basename + ".o")


def check_results(results, log=None, print_w=None):
  """
  Check the result of system commands exectution.

  Parameters
  ----------
  results : list
    list of exectuions results
  log : {False}
    bool for activate errors log saving
  print_w : {None}
    function for printing emphized warning message
  """
  if print_w is None:
    print_wm = print_fake
  else:
    print_wm = print_w

  if not all(v[1] == '' for v in results):
    for result in results:
      if result[1] != '':
        print_wm(result[1])
  if not all(v[0] == 0 for v in results):
    for result in results:
      if result[0] != 0:
        sys.stderr.write(result[1])
        if log:
          with open(log, 'a') as logerror:
            logerror.writelines(result[1])
    sys.exit(1)


def safe_mkdir(directory, print_w=None):
  """
  Create directory via safe checkings.

  Parameters
  ----------
  directory: str
    path of the directory to be created
  print_w : {None}
    function for printing emphized warning message
  """
  if print_w is None:
    print_wm = print_fake
  else:
    print_wm = print_w

  if os.path.exists(directory):
    if not os.path.isdir(directory):
      print_wm('Error: cannot create directory "' + directory + '", a file with the same name already exists!')
      sys.exit(1)
  else:
    os.makedirs(directory)
  return



"""
Dependency.py, module definition of Dependency class.

This is a class designed for handling file dependency.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
# from builtins import object
import os


class Dependency(object):

  """Dependency is an object that handles a single file dependency, its attributes and methods."""

  def __init__(self, dtype="", name="", dfile=""):
    """
    Parameters
    ----------
    dtype : {""}
      type of dependency: "module" or "include" type
    name : {""}
      name of dependency: module name for "use" type or file name for include type
    file : {""}
      file name containing module in the case of "use" type

    Attributes
    ----------
    """
    self.type = dtype
    self.name = name
    self.file = dfile

  def __str__(self):
    string = []
    string.append('\n  Type: ' + str(self.type))
    string.append('\n  Name: ' + str(self.name))
    string.append('\n  File: ' + str(self.file))
    return ''.join(string)

  def printf(self):
    """Method for printing dependency data."""
    print(self)
    return

  def exist(self):
    """Method for checking the existance of a dependency file."""
    return os.path.exists(self.file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: /home/stefano/python/FoBiS/src/main/python/fobis/Fobos.py
# Author: Stefano Zaghi <stefano.zaghi@gmail.com>
# Date: 28.08.2017
# Last Modified Date: 28.08.2017
# Last Modified By: Stefano Zaghi <stefano.zaghi@gmail.com>
"""
fobos.py, module definition of fobos class.
This is a class aimed at fobos file handling.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from __future__ import print_function
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import unicode_literals
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
# from builtins import object
try:
  import configparser as configparser
except ImportError:
  import configparser
from copy import deepcopy
import os
import re
import sys
from .utils import check_results, print_fake, syswork


class Fobos(object):
  """
  Fobos is an object that handles fobos file, its attributes and methods.
  """
  def __init__(self, cliargs, print_n=None, print_w=None):
    """
    Parameters
    ----------
    cliargs : argparse object
    print_n : {None}
      function for printing normal message
    print_w : {None}
      function for printing emphized warning message
    """
    if print_n is None:
      self.print_n = print_fake
    else:
      self.print_n = print_n

    if print_w is None:
      self.print_w = print_fake
    else:
      self.print_w = print_w

    self.fobos = None
    self.mode = None
    self.local_variables = {}
    if cliargs.fobos:
      filename = cliargs.fobos
    else:
      filename = 'fobos'
    if os.path.exists(filename):
      self.fobos = configparser.RawConfigParser()
      if not cliargs.fobos_case_insensitive:
        self.fobos.optionxform = str  # case sensitive
      self.fobos.read(filename)
      self._set_cliargs(cliargs=cliargs)
    return

  def _check_mode(self, mode):
    """
    Function for checking the presence of the selected mode into the set defined inside the fobos.

    Parameters
    ----------
    mode : str
      name of the selcted mode
    """
    if self.fobos:
      if self.fobos.has_option('modes', 'modes'):
        if mode in self.fobos.get('modes', 'modes'):
          self.mode = mode
        else:
          self.print_w('Error: the mode "' + mode + '" is not defined into the fobos file.')
          self.modes_list()
          sys.exit(1)
      else:
        self.print_w('Error: fobos file has not "modes" section.')
        sys.exit(1)
    return

  def _set_mode(self, mode=None):
    """
    Function for setting the selected mode.

    Parameters
    ----------
    mode : {None}
      selected mode
    """
    if self.fobos:
      if mode:
        self._check_mode(mode=mode)
      else:
        if self.fobos.has_option('modes', 'modes'):
          self.mode = self.fobos.get('modes', 'modes').split()[0]  # first mode selected
        else:
          if self.fobos.has_section('default'):
            self.mode = 'default'
          else:
            self.print_w('Warning: fobos file has not "modes" section neither "default" one')
    return

  def _check_template(self):
    """
    Check and apply template sections.

    Each mode may specify one or more template names as a space-separated
    list.  Templates are applied left-to-right: the mode itself always
    wins, and earlier templates take precedence over later ones.
    Template-of-template inheritance is expanded depth-first.
    Circular references are detected and cause an error.
    """
    if not self.fobos:
      return

    def _template_names(section):
      if self.fobos.has_option(section, 'template'):
        return self.fobos.get(section, 'template').split()
      return []

    def _resolve(section, visiting):
      """Return ordered list of template names to apply for section."""
      if section in visiting:
        self.print_w('Error: circular template reference detected involving "' + section + '"')
        sys.exit(1)
      visiting = visiting | {section}
      resolved = []
      for name in _template_names(section):
        if not self.fobos.has_section(name):
          self.print_w('Error: mode "' + section + '" uses template "' + name + '" that is NOT defined')
          sys.exit(1)
        if name not in resolved:
          resolved.append(name)
        for sub in _resolve(name, visiting):
          if sub not in resolved:
            resolved.append(sub)
      return resolved

    for section in self.fobos.sections():
      if not self.fobos.has_option(section, 'template'):
        continue
      for template in _resolve(section, set()):
        for item in self.fobos.items(template):
          if item[0] == 'template':
            continue
          if not self.fobos.has_option(section, item[0]):
            self.fobos.set(section, item[0], item[1])
    return

  def _get_local_variables(self):
    """
    Get the definition of local variables defined into any sections (modes).
    """
    if self.fobos:
      for section in self.fobos.sections():
        for item in self.fobos.items(section):
          if item[0].startswith('$'):
            self.local_variables[item[0]] = item[1].replace('\n', ' ')
    return

  def _substitute_local_variables_mode(self):
    """
    Substitute the definition of local variables defined into the mode (section) selected.
    """
    if self.fobos and self.mode:
      self._substitute_local_variables_section(section=self.mode)
    return

  def _substitute_local_variables_section(self, section):
    """
    Substitute the definition of local variables defined into a section.
    """
    if self.fobos:
      if self.fobos.has_section(section):
        for item in self.fobos.items(section):
          item_val = item[1]
          for key, value in list(self.local_variables.items()):
            item_val = re.sub(re.escape(key), value, item_val)
            # item_val = re.sub(r"(?!" + re.escape(key) + r"[aZ_-])\s*" + re.escape(key) + r"\s*", value, item_val)
          self.fobos.set(section, item[0], item_val)
    return

  def _check_local_variables(self):
    """
    Get and substitute the definition of local variables defined into any sections (modes).
    """
    if self.fobos:
      self._get_local_variables()
      if len(self.local_variables) > 0:
        self._substitute_local_variables_mode()
    return

  def _set_cliargs_attributes(self, cliargs, cliargs_dict):
    """
    Set attributes of cliargs from fobos options.

    Parameters
    ----------
    cliargs : argparse object
    cliargs_dict : argparse object attributes dictionary
    """
    if self.mode:
      for item in self.fobos.items(self.mode):
        if item[0] in cliargs_dict:
          if isinstance(cliargs_dict[item[0]], bool):
            setattr(cliargs, item[0], self.fobos.getboolean(self.mode, item[0]))
          elif isinstance(cliargs_dict[item[0]], int):
            setattr(cliargs, item[0], int(item[1]))
          elif isinstance(cliargs_dict[item[0]], list):
            setattr(cliargs, item[0], item[1].split())
          else:
            setattr(cliargs, item[0], item[1])
    return

  @staticmethod
  def _check_cliargs_cflags(cliargs, cliargs_dict):
    """
    Method for setting attribute of cliargs.

    Parameters
    ----------
    cliargs : argparse object
    cliargs_dict : argparse object attributes dictionary
    """
    for item in cliargs_dict:
      if item in ['cflags', 'lflags', 'preproc']:
        val_cli = cliargs_dict[item]
        val_fobos = getattr(cliargs, item)
        if item == 'cflags':
          if val_cli == '-c':
            match = re.search(r'(-c\s+|-c$)', val_fobos)
            if match:
              val_cli = ''  # avoid multiple -c flags
        if val_fobos and val_cli:
          setattr(cliargs, item, val_fobos + ' ' + val_cli)
    return

  def _set_cliargs(self, cliargs):
    """
    Set cliargs from fobos options.

    Parameters
    ----------
    cliargs : argparse object
    """
    if self.fobos:
      cliargs_dict = deepcopy(cliargs.__dict__)
      self._set_mode(mode=cliargs.mode)
      self._check_template()
      self._check_local_variables()
      self._set_cliargs_attributes(cliargs=cliargs, cliargs_dict=cliargs_dict)
      self._check_cliargs_cflags(cliargs=cliargs, cliargs_dict=cliargs_dict)
    return

  def get(self, option, mode=None, toprint=True):
    """
    Get options defined into the fobos file.

    Parameters
    ----------
    option : str
      option name
    mode : {None}
      eventual mode name
    toprint : {True}
      return of the value: if toprint==False the value is return otherwise is printed to stdout
    """
    value = ''
    if self.fobos:
      self._set_mode(mode=mode)
      if self.fobos.has_option(self.mode, option):
        value = self.fobos.get(self.mode, option)
    if toprint:
      # self.print_w(value)
      print(value)
      return
    else:
      return value

  def get_output_name(self, mode=None, toprint=True):
    """
    Method for building the output name accordingly to the fobos options.

    Parameters
    ----------
    mode : {None}
      eventual mode name
    toprint : {True}
      return of the value: if toprint==False the value is return otherwise is printed to stdout
    """
    output = ''
    build_dir = self.get(option='build_dir', mode=mode, toprint=False)
    mklib = self.get(option='mklib', mode=mode, toprint=False)
    if self.fobos:
      self._set_mode(mode=mode)
      if self.fobos.has_option(self.mode, 'output'):
        output = self.fobos.get(self.mode, 'output')
        output = os.path.normpath(os.path.join(build_dir, output))
      elif self.fobos.has_option(self.mode, 'target'):
        output = self.fobos.get(self.mode, 'target')
        output = os.path.splitext(os.path.basename(output))[0]
        if mklib.lower() == 'shared':
          output = output + '.so'
        elif mklib.lower() == 'static':
          output = output + '.a'
        output = os.path.normpath(os.path.join(build_dir, output))
    if toprint:
      # self.print_w(output)
      print(output)
      return
    else:
      return output

  def modes_list(self):
    """List defined modes."""
    if self.fobos:
      self.print_n('The fobos file defines the following modes:')
      if self.fobos.has_option('modes', 'modes'):
        modes = self.fobos.get('modes', 'modes').split()
        for mode in modes:
          if self.fobos.has_section(mode):
            if self.fobos.has_option(mode, 'help'):
              helpmsg = self.fobos.get(mode, 'help')
            else:
              helpmsg = ''
            self.print_n('  - "' + mode + '" ' + helpmsg)
      else:
        self.print_w('Error: no modes are defined into the fobos file!')
        sys.exit(1)
    sys.exit(0)
    return

  @staticmethod
  def print_template(cliargs):
    """
    Print fobos template.

    Parameters
    ----------
    cliargs : argparse object
    """
    print("[default]")
    for argument in vars(cliargs):
      attribute = getattr(cliargs, argument)
      if isinstance(attribute, list):
        attribute = ' '.join(attribute)
      print(str(argument) + " = " + str(attribute))

  def get_project_info(self):
    """
    Parse [project] section and return project metadata.

    Returns
    -------
    dict
      dict with keys 'name' (str), 'authors' (list of str),
      'version' (str, raw value as written in fobos — not resolved),
      'summary' (str), 'repository' (str), 'website' (str), and 'email' (str).
      All values are empty/empty-list if the section or option is absent.
    """
    info = {'name': '', 'authors': [], 'version': '', 'summary': '', 'repository': '', 'website': '', 'email': ''}
    if self.fobos and self.fobos.has_section('project'):
      if self.fobos.has_option('project', 'name'):
        info['name'] = self.fobos.get('project', 'name').strip()
      if self.fobos.has_option('project', 'authors'):
        raw = self.fobos.get('project', 'authors')
        info['authors'] = [a.strip() for a in raw.splitlines() if a.strip()]
      if self.fobos.has_option('project', 'version'):
        info['version'] = self.fobos.get('project', 'version').strip()
      if self.fobos.has_option('project', 'summary'):
        info['summary'] = self.fobos.get('project', 'summary').strip()
      if self.fobos.has_option('project', 'repository'):
        info['repository'] = self.fobos.get('project', 'repository').strip()
      if self.fobos.has_option('project', 'website'):
        info['website'] = self.fobos.get('project', 'website').strip()
      if self.fobos.has_option('project', 'email'):
        info['email'] = self.fobos.get('project', 'email').strip()
    return info

  def get_version(self):
    """
    Resolve the project version from [project] and/or git tags.

    Resolution steps
    ----------------
    1. Read 'version' from [project] in fobos.  If the value is a
       file path (relative to the git repository root), the version
       string is read from that file.
    2. Query the most recent git tag via ``git describe --tags --abbrev=0``.
    3. If both sources provide a version and they disagree, emit a
       warning with a suggested fix.
    4. Return the fobos version when available; fall back to the git
       tag; return '' when neither source is determinable.

    Returns
    -------
    str
      Resolved version string, or '' if not determinable.
    """
    fobos_version = ''
    if self.fobos and self.fobos.has_section('project'):
      if self.fobos.has_option('project', 'version'):
        raw = self.fobos.get('project', 'version').strip()
        # try to resolve as a file path relative to the git repo root
        git_root_result = syswork('git rev-parse --show-toplevel')
        if git_root_result[0] == 0:
          candidate = os.path.join(git_root_result[1].strip(), raw)
          if os.path.isfile(candidate):
            with open(candidate) as ver_file:
              fobos_version = ver_file.read().strip()
          else:
            fobos_version = raw
        else:
          fobos_version = raw  # not inside a git repo; treat as literal
        if fobos_version and not fobos_version.startswith('v'):
          fobos_version = 'v' + fobos_version

    # query the most recent git tag
    git_version = ''
    git_result = syswork('git describe --tags --abbrev=0')
    if git_result[0] == 0:
      git_version = git_result[1].strip()

    # warn on mismatch
    if fobos_version and git_version and fobos_version != git_version:
      git_version_v = git_version if git_version.startswith('v') else 'v' + git_version
      self.print_w('Warning: project version mismatch!')
      self.print_w('  fobos [project] version : ' + fobos_version)
      self.print_w('  git tag version         : ' + git_version)
      self.print_w('  To fix, either:')
      self.print_w('    - update fobos: set  version = ' + git_version_v + '  under [project]')
      self.print_w('    - create a matching tag: git tag ' + fobos_version + ' && git push --tags')

    return fobos_version or git_version

  def get_dependencies(self):
    """
    Parse [dependencies] section and return dict of {name: spec_string}.

    Returns
    -------
    dict
      mapping of dependency name to its spec string, or empty dict if no section
    """
    deps = {}
    if self.fobos and self.fobos.has_section('dependencies'):
      for name, spec in self.fobos.items('dependencies'):
        if name == 'deps_dir':
          continue
        deps[name] = spec
    return deps

  def get_deps_dir(self, default='.fobis_deps'):
    """
    Read deps_dir from [dependencies] section of fobos.

    Parameters
    ----------
    default : str
      value returned when the option is absent [default: '.fobis_deps']

    Returns
    -------
    str
      deps_dir value from fobos, or default if not set
    """
    if self.fobos and self.fobos.has_section('dependencies'):
      if self.fobos.has_option('dependencies', 'deps_dir'):
        return self.fobos.get('dependencies', 'deps_dir').strip()
    return default

  def rules_list(self, quiet=False):
    """
    Function for listing defined rules.

    Parameters
    ----------
    quiet : {False}
      less verbose outputs than default
    """
    if self.fobos:
      self.print_n('The fobos file defines the following rules:')
      for rule in self.fobos.sections():
        if rule.startswith('rule-'):
          if self.fobos.has_option(rule, 'help'):
            helpmsg = self.fobos.get(rule, 'help')
          else:
            helpmsg = ''
          self.print_n('  - "' + rule.split('rule-')[1] + '" ' + helpmsg)
          if self.fobos.has_option(rule, 'quiet'):
            quiet = self.fobos.getboolean(rule, 'quiet')
          for rul in self.fobos.options(rule):
            if rul.startswith('rule'):
              if not quiet:
                self.print_n('       Command => ' + self.fobos.get(rule, rul))
    sys.exit(0)
    return

  def rule_execute(self, rule, quiet=False, log=False):
    """
    Function for executing selected rule.

    Parameters
    ----------
    rule : str
      rule name
    quiet : {False}
      less verbose outputs than default
    log : {False}
      bool for activate errors log saving
    """
    if self.fobos:
      self.print_n('Executing rule "' + rule + '"')
      rule_name = 'rule-' + rule
      if self.fobos.has_section(rule_name):
        self._get_local_variables()
        self._substitute_local_variables_section(section=rule_name)
        results = []
        quiet = False
        log = False
        if self.fobos.has_option(rule_name, 'quiet'):
          quiet = self.fobos.getboolean(rule_name, 'quiet')
        if self.fobos.has_option(rule_name, 'log'):
          log = self.fobos.getboolean(rule_name, 'log')
        for rul in self.fobos.options(rule_name):
          if rul.startswith('rule'):
            if not quiet:
              self.print_n('   Command => ' + self.fobos.get(rule_name, rul))
            result = syswork(self.fobos.get(rule_name, rul))
            results.append(result)
        if log:
          check_results(results=results, log='rules_errors.log', print_w=self.print_w)
        else:
          check_results(results=results, print_w=self.print_w)
      else:
        self.print_w('Error: the rule "' + rule + '" is not defined into the fobos file. Defined rules are:')
        self.rules_list(quiet=quiet)
        sys.exit(1)
    return



"""
cli_parser.py, FoBiS.py CLI definition built with Typer.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
import argparse
import re
from typing import List, Optional
from typing_extensions import Annotated
import typer

# ---------------------------------------------------------------------------
# Module-level constants (imported by other modules — do not rename)
# ---------------------------------------------------------------------------
__extensions_inc__ = [".inc", ".INC", ".h", ".H"]
__extensions_old__ = [".f", ".F", ".for", ".FOR", ".fpp", ".FPP", ".fortran", ".f77", ".F77"]
__extensions_modern__ = [".f90", ".F90", ".f95", ".F95", ".f03", ".F03", ".f08", ".F08", ".f2k", ".F2K"]
__extensions_parsed__ = __extensions_inc__ + __extensions_old__ + __extensions_modern__
__compiler_supported__ = ('gnu', 'intel', 'intel_nextgen', 'g95', 'opencoarrays-gnu', 'pgi', 'ibm', 'nag', 'nvfortran', 'amd', 'custom')

# ---------------------------------------------------------------------------
# Argument normaliser — preserves backward compat with argparse-style options
# ---------------------------------------------------------------------------
_MULTI_CHAR_OPT = re.compile(r'^-[A-Za-z][A-Za-z0-9_-]+$')


def _normalize_args(args):
  """
  Normalise FoBiS legacy single-dash long options for Click/Typer.

  Rules applied to each token:
  - Single-dash multi-char option (-compiler, -mode, -get_output_name)
    → double-dash with underscores turned to hyphens (--compiler, --mode, --get-output-name)
  - Double-dash option with underscores (--build_dir, --cflags_heritage)
    → double-dash with hyphens (--build-dir, --cflags-heritage)
  - Single-char short options (-f, -m, -q), values, and negative numbers
    are left unchanged.
  """
  result = []
  for arg in args:
    if _MULTI_CHAR_OPT.match(arg):
      # -compiler → --compiler ; -get_output_name → --get-output-name
      result.append('--' + arg[1:].replace('_', '-'))
    elif arg.startswith('--') and len(arg) > 2:
      # --build_dir → --build-dir  (handle optional = form too)
      if '=' in arg:
        opt, val = arg[2:].split('=', 1)
        result.append('--' + opt.replace('_', '-') + '=' + val)
      else:
        result.append('--' + arg[2:].replace('_', '-'))
    else:
      result.append(arg)
  return result

# ---------------------------------------------------------------------------
# Typer application
# ---------------------------------------------------------------------------
app = typer.Typer(
  name='FoBiS.py',
  help='a Fortran Building System for poor men',
  no_args_is_help=True,
  add_completion=True,
  rich_markup_mode=None,
)


def _version_callback(value: bool):
  if value:
    # lazy import avoids circular dependency
    from .FoBiSConfig import __version__, __appname__
    typer.echo(f'{__appname__} {__version__}')
    raise typer.Exit()


@app.callback()
def _app_callback(
  ctx: typer.Context,
  version: Annotated[bool, typer.Option(
    '--version', '-v',
    help='Show version and exit.',
    callback=_version_callback,
    is_eager=True,
  )] = False,
):
  ctx.ensure_object(dict)

# ---------------------------------------------------------------------------
# Autocompletion callbacks
# ---------------------------------------------------------------------------

def _complete_compiler(incomplete: str):
  return [c for c in __compiler_supported__ if c.startswith(incomplete.lower())]


def _complete_mklib(incomplete: str):
  return [m for m in ('static', 'shared') if m.startswith(incomplete)]


def _complete_extensions(incomplete: str):
  return [e for e in __extensions_parsed__ if e.startswith(incomplete)]


def _complete_doctests_preprocessor(incomplete: str):
  return [p for p in ('cpp', 'fpp') if p.startswith(incomplete)]


def _complete_fobos_mode(ctx: typer.Context, incomplete: str):
  import configparser
  import os
  fobos_path = ctx.params.get('fobos') or 'fobos'
  if not os.path.exists(fobos_path):
    return []
  cp = configparser.RawConfigParser()
  cp.read(fobos_path)
  if cp.has_option('modes', 'modes'):
    modes = [m.strip() for m in cp.get('modes', 'modes').split()]
    return [m for m in modes if m.startswith(incomplete)]
  return [s for s in cp.sections() if s.startswith(incomplete) and s not in ('modes', 'rules', 'dependencies', 'project')]

# ---------------------------------------------------------------------------
# Namespace factory
# ---------------------------------------------------------------------------

def _ns(**kwargs) -> argparse.Namespace:
  """Build an argparse.Namespace — preserves the duck-type expected by all downstream code."""
  return argparse.Namespace(**kwargs)

# ---------------------------------------------------------------------------
# build command
# ---------------------------------------------------------------------------

@app.command('build')
def cmd_build(
  ctx: typer.Context,
  # fobos group
  fobos:                    Annotated[Optional[str],       typer.Option('--fobos', '-f',           help='Specify a "fobos" file named differently from "fobos"')] = None,
  fobos_case_insensitive:   Annotated[bool,                typer.Option('--fci',                   help='Assume fobos inputs as case insensitive')] = False,
  mode:                     Annotated[Optional[str],       typer.Option('--mode',                  help='Select a mode defined into a fobos file', autocompletion=_complete_fobos_mode)] = None,
  lmodes:                   Annotated[bool,                typer.Option('--lmodes',                help='List the modes defined into a fobos file')] = False,
  print_fobos_template:     Annotated[bool,                typer.Option('--print-fobos-template',  help='Print a comprehensive fobos template')] = False,
  # compiler group
  compiler:                 Annotated[str,                 typer.Option('--compiler',              help='Compiler used (case insensitive, default gnu)', autocompletion=_complete_compiler)] = 'gnu',
  fc:                       Annotated[Optional[str],       typer.Option('--fc',                    help='Fortran compiler statement (for --compiler custom)')] = None,
  cflags:                   Annotated[Optional[str],       typer.Option('--cflags',                help='Compile flags')] = None,
  lflags:                   Annotated[Optional[str],       typer.Option('--lflags',                help='Link flags')] = None,
  modsw:                    Annotated[Optional[str],       typer.Option('--modsw',                 help='Module search path switch (for --compiler custom)')] = None,
  mpi:                      Annotated[bool,                typer.Option('--mpi',                   help='Use MPI enabled version of compiler')] = False,
  openmp:                   Annotated[bool,                typer.Option('--openmp',                help='Use OpenMP pragmas')] = False,
  openmp_offload:           Annotated[bool,                typer.Option('--openmp-offload',        help='Use OpenMP Offload pragmas')] = False,
  coarray:                  Annotated[bool,                typer.Option('--coarray',               help='Use coarrays')] = False,
  coverage:                 Annotated[bool,                typer.Option('--coverage',              help='Instrument for coverage analysis')] = False,
  profile:                  Annotated[bool,                typer.Option('--profile',               help='Instrument for profiling analysis')] = False,
  mklib:                    Annotated[Optional[str],       typer.Option('--mklib',                 help='Build library instead of program (static|shared)', autocompletion=_complete_mklib)] = None,
  ar:                       Annotated[str,                 typer.Option('--ar',                    help='Archiver executable for static libraries [default: ar]')] = 'ar',
  arflags:                  Annotated[str,                 typer.Option('--arflags',               help='Archiver flags for static libraries [default: -rcs]')] = '-rcs',
  ranlib:                   Annotated[str,                 typer.Option('--ranlib',                help='Ranlib executable; empty string to skip [default: ranlib]')] = 'ranlib',
  cflags_heritage:          Annotated[bool,                typer.Option('--cflags-heritage', '--ch', help='Store cflags for heritage checking; re-compile all if they change')] = False,
  track_build:              Annotated[bool,                typer.Option('--track-build', '--tb',   help='Store build info for the install command')] = False,
  # directories group
  src:                      Annotated[Optional[List[str]], typer.Option('--src', '-s',             help='Root-directory of source files [default: ./]')] = None,
  build_dir:                Annotated[str,                 typer.Option('--build-dir', '--dbld',   help='Directory containing built objects [default: ./]')] = './',
  obj_dir:                  Annotated[str,                 typer.Option('--obj-dir', '--dobj',     help='Directory containing compiled objects [default: ./obj/]')] = './obj/',
  mod_dir:                  Annotated[str,                 typer.Option('--mod-dir', '--dmod',     help='Directory containing .mod files [default: ./mod/]')] = './mod/',
  lib_dir:                  Annotated[Optional[List[str]], typer.Option('--lib-dir', '--dlib',     help='Directories searched for libraries')] = None,
  include:                  Annotated[Optional[List[str]], typer.Option('--include', '-i',         help='Directories for searching included files')] = None,
  exclude_dirs:             Annotated[Optional[List[str]], typer.Option('--exclude-dirs', '--ed',  help='Exclude directories from the building process')] = None,
  disable_recursive_search: Annotated[bool,                typer.Option('--disable-recursive-search', '--drs', help='Disable recursive search inside directories')] = False,
  # files group
  target:                   Annotated[Optional[str],       typer.Option('--target', '-t',          help='Specify a target file [default: all programs found]')] = None,
  output:                   Annotated[Optional[str],       typer.Option('--output', '-o',          help='Output file name (used with --target)')] = None,
  exclude:                  Annotated[Optional[List[str]], typer.Option('--exclude', '-e',         help='Exclude files from the building process')] = None,
  libs:                     Annotated[Optional[List[str]], typer.Option('--libs',                  help='External libraries with full paths')] = None,
  vlibs:                    Annotated[Optional[List[str]], typer.Option('--vlibs',                 help='Volatile external libraries with full paths')] = None,
  ext_libs:                 Annotated[Optional[List[str]], typer.Option('--ext-libs',              help='External libraries in compiler path')] = None,
  ext_vlibs:                Annotated[Optional[List[str]], typer.Option('--ext-vlibs',             help='Volatile external libraries in compiler path')] = None,
  dependon:                 Annotated[Optional[List[str]], typer.Option('--dependon',              help='Interdependent external fobos files for interdependent building')] = None,
  inc:                      Annotated[Optional[List[str]], typer.Option('--inc',                   help='Extensions for include files', autocompletion=_complete_extensions)] = None,
  extensions:               Annotated[Optional[List[str]], typer.Option('--extensions',            help='Extensions of parsed files', autocompletion=_complete_extensions)] = None,
  build_all:                Annotated[bool,                typer.Option('--build-all',             help='Build all sources parsed')] = False,
  # preprocessor group
  preprocessor:             Annotated[Optional[str],       typer.Option('--preprocessor',          help='Preprocessor name for pre-processing sources (e.g. PreForM.py)')] = None,
  preproc:                  Annotated[Optional[str],       typer.Option('--preproc',               help='Preprocessor flags for the main compiler')] = None,
  preprocessor_args:        Annotated[str,                 typer.Option('--preprocessor-args', '--app', help='Preprocessor-specific flags')] = '',
  preprocessor_no_o:        Annotated[bool,                typer.Option('--preprocessor-no-o', '--npp', help='Do not add -o in front of output in preprocessor command')] = False,
  preprocessor_dir:         Annotated[Optional[str],       typer.Option('--preprocessor-dir', '--dpp', help='Directory for preprocessed files')] = None,
  preprocessor_ext:         Annotated[Optional[List[str]], typer.Option('--preprocessor-ext', '--epp', help='File extensions to preprocess')] = None,
  # fancy group
  force_compile:            Annotated[bool,                typer.Option('--force-compile',         help='Force to (re-)compile all')] = False,
  colors:                   Annotated[bool,                typer.Option('--colors',                help='Activate colors in shell prints')] = False,
  log:                      Annotated[bool,                typer.Option('--log', '-l',             help='Activate log file creation')] = False,
  graph:                    Annotated[bool,                typer.Option('--graph',                 help='Generate a dependencies graph via graphviz')] = False,
  quiet:                    Annotated[bool,                typer.Option('--quiet', '-q',           help='Less verbose output')] = False,
  verbose:                  Annotated[bool,                typer.Option('--verbose',               help='Extremely verbose output for debugging')] = False,
  jobs:                     Annotated[int,                 typer.Option('--jobs', '-j',            help='Number of concurrent compilation jobs [default: 1]')] = 1,
  makefile:                 Annotated[Optional[str],       typer.Option('--makefile', '-m',        help='Generate a GNU Makefile for building the project')] = None,
):
  """Build all programs found or specific target(s)."""
  ctx.ensure_object(dict)
  ctx.obj['cliargs'] = _ns(
    which='build',
    fobos=fobos, fobos_case_insensitive=fobos_case_insensitive,
    mode=mode, lmodes=lmodes, print_fobos_template=print_fobos_template,
    compiler=compiler.lower(), fc=fc, cflags=cflags, lflags=lflags, modsw=modsw,
    mpi=mpi, openmp=openmp, openmp_offload=openmp_offload,
    coarray=coarray, coverage=coverage, profile=profile,
    mklib=mklib, ar=ar, arflags=arflags, ranlib=ranlib,
    cflags_heritage=cflags_heritage, track_build=track_build,
    src=src or ['./'],
    build_dir=build_dir, obj_dir=obj_dir, mod_dir=mod_dir,
    lib_dir=lib_dir or [], include=include or [],
    exclude_dirs=exclude_dirs or [],
    disable_recursive_search=disable_recursive_search,
    target=target, output=output,
    exclude=exclude or [], libs=libs or [], vlibs=vlibs or [],
    ext_libs=ext_libs or [], ext_vlibs=ext_vlibs or [],
    dependon=dependon or [],
    inc=inc or list(__extensions_inc__),
    extensions=extensions or list(__extensions_parsed__),
    build_all=build_all,
    preprocessor=preprocessor, preproc=preproc,
    preprocessor_args=preprocessor_args,
    preprocessor_no_o=preprocessor_no_o,
    preprocessor_dir=preprocessor_dir,
    preprocessor_ext=preprocessor_ext or [],
    force_compile=force_compile, colors=colors, log=log, graph=graph,
    quiet=quiet, verbose=verbose, jobs=jobs, makefile=makefile,
  )

# ---------------------------------------------------------------------------
# clean command
# ---------------------------------------------------------------------------

@app.command('clean')
def cmd_clean(
  ctx: typer.Context,
  # fobos group
  fobos:                    Annotated[Optional[str],       typer.Option('--fobos', '-f',           help='Specify a "fobos" file named differently from "fobos"')] = None,
  fobos_case_insensitive:   Annotated[bool,                typer.Option('--fci',                   help='Assume fobos inputs as case insensitive')] = False,
  mode:                     Annotated[Optional[str],       typer.Option('--mode',                  help='Select a mode defined into a fobos file', autocompletion=_complete_fobos_mode)] = None,
  lmodes:                   Annotated[bool,                typer.Option('--lmodes',                help='List the modes defined into a fobos file')] = False,
  print_fobos_template:     Annotated[bool,                typer.Option('--print-fobos-template',  help='Print a comprehensive fobos template')] = False,
  # clean-specific (were part of compiler group in argparse)
  only_obj:                 Annotated[bool,                typer.Option('--only-obj',              help='Clean only compiled objects and not built targets')] = False,
  only_target:              Annotated[bool,                typer.Option('--only-target',           help='Clean only built targets and not compiled objects')] = False,
  # compiler group
  compiler:                 Annotated[str,                 typer.Option('--compiler',              help='Compiler used (case insensitive, default gnu)', autocompletion=_complete_compiler)] = 'gnu',
  fc:                       Annotated[Optional[str],       typer.Option('--fc',                    help='Fortran compiler statement (for --compiler custom)')] = None,
  cflags:                   Annotated[Optional[str],       typer.Option('--cflags',                help='Compile flags')] = None,
  lflags:                   Annotated[Optional[str],       typer.Option('--lflags',                help='Link flags')] = None,
  modsw:                    Annotated[Optional[str],       typer.Option('--modsw',                 help='Module search path switch (for --compiler custom)')] = None,
  mpi:                      Annotated[bool,                typer.Option('--mpi',                   help='Use MPI enabled version of compiler')] = False,
  openmp:                   Annotated[bool,                typer.Option('--openmp',                help='Use OpenMP pragmas')] = False,
  openmp_offload:           Annotated[bool,                typer.Option('--openmp-offload',        help='Use OpenMP Offload pragmas')] = False,
  coarray:                  Annotated[bool,                typer.Option('--coarray',               help='Use coarrays')] = False,
  coverage:                 Annotated[bool,                typer.Option('--coverage',              help='Instrument for coverage analysis')] = False,
  profile:                  Annotated[bool,                typer.Option('--profile',               help='Instrument for profiling analysis')] = False,
  mklib:                    Annotated[Optional[str],       typer.Option('--mklib',                 help='Build library instead of program (static|shared)', autocompletion=_complete_mklib)] = None,
  ar:                       Annotated[str,                 typer.Option('--ar',                    help='Archiver executable for static libraries [default: ar]')] = 'ar',
  arflags:                  Annotated[str,                 typer.Option('--arflags',               help='Archiver flags for static libraries [default: -rcs]')] = '-rcs',
  ranlib:                   Annotated[str,                 typer.Option('--ranlib',                help='Ranlib executable; empty string to skip [default: ranlib]')] = 'ranlib',
  cflags_heritage:          Annotated[bool,                typer.Option('--cflags-heritage', '--ch', help='Store cflags for heritage checking')] = False,
  track_build:              Annotated[bool,                typer.Option('--track-build', '--tb',   help='Store build info for the install command')] = False,
  # directories group
  src:                      Annotated[Optional[List[str]], typer.Option('--src', '-s',             help='Root-directory of source files [default: ./]')] = None,
  build_dir:                Annotated[str,                 typer.Option('--build-dir', '--dbld',   help='Directory containing built objects [default: ./]')] = './',
  obj_dir:                  Annotated[str,                 typer.Option('--obj-dir', '--dobj',     help='Directory containing compiled objects [default: ./obj/]')] = './obj/',
  mod_dir:                  Annotated[str,                 typer.Option('--mod-dir', '--dmod',     help='Directory containing .mod files [default: ./mod/]')] = './mod/',
  lib_dir:                  Annotated[Optional[List[str]], typer.Option('--lib-dir', '--dlib',     help='Directories searched for libraries')] = None,
  include:                  Annotated[Optional[List[str]], typer.Option('--include', '-i',         help='Directories for searching included files')] = None,
  exclude_dirs:             Annotated[Optional[List[str]], typer.Option('--exclude-dirs', '--ed',  help='Exclude directories from the building process')] = None,
  disable_recursive_search: Annotated[bool,                typer.Option('--disable-recursive-search', '--drs', help='Disable recursive search inside directories')] = False,
  # files group
  target:                   Annotated[Optional[str],       typer.Option('--target', '-t',          help='Specify a target file')] = None,
  output:                   Annotated[Optional[str],       typer.Option('--output', '-o',          help='Output file name (used with --target)')] = None,
  exclude:                  Annotated[Optional[List[str]], typer.Option('--exclude', '-e',         help='Exclude files from the building process')] = None,
  libs:                     Annotated[Optional[List[str]], typer.Option('--libs',                  help='External libraries with full paths')] = None,
  vlibs:                    Annotated[Optional[List[str]], typer.Option('--vlibs',                 help='Volatile external libraries with full paths')] = None,
  ext_libs:                 Annotated[Optional[List[str]], typer.Option('--ext-libs',              help='External libraries in compiler path')] = None,
  ext_vlibs:                Annotated[Optional[List[str]], typer.Option('--ext-vlibs',             help='Volatile external libraries in compiler path')] = None,
  dependon:                 Annotated[Optional[List[str]], typer.Option('--dependon',              help='Interdependent external fobos files')] = None,
  inc:                      Annotated[Optional[List[str]], typer.Option('--inc',                   help='Extensions for include files')] = None,
  extensions:               Annotated[Optional[List[str]], typer.Option('--extensions',            help='Extensions of parsed files')] = None,
  build_all:                Annotated[bool,                typer.Option('--build-all',             help='Build all sources parsed')] = False,
  # fancy group
  force_compile:            Annotated[bool,                typer.Option('--force-compile',         help='Force to (re-)compile all')] = False,
  colors:                   Annotated[bool,                typer.Option('--colors',                help='Activate colors in shell prints')] = False,
  log:                      Annotated[bool,                typer.Option('--log', '-l',             help='Activate log file creation')] = False,
  graph:                    Annotated[bool,                typer.Option('--graph',                 help='Generate a dependencies graph via graphviz')] = False,
  quiet:                    Annotated[bool,                typer.Option('--quiet', '-q',           help='Less verbose output')] = False,
  verbose:                  Annotated[bool,                typer.Option('--verbose',               help='Extremely verbose output for debugging')] = False,
  jobs:                     Annotated[int,                 typer.Option('--jobs', '-j',            help='Number of concurrent compilation jobs [default: 1]')] = 1,
  makefile:                 Annotated[Optional[str],       typer.Option('--makefile', '-m',        help='Generate a GNU Makefile for building the project')] = None,
):
  """Clean project: remove all OBJs and MODs files. Use carefully."""
  ctx.ensure_object(dict)
  if only_obj and only_target:
    typer.echo('Error: --only-obj and --only-target are mutually exclusive.', err=True)
    raise typer.Exit(1)
  ctx.obj['cliargs'] = _ns(
    which='clean',
    fobos=fobos, fobos_case_insensitive=fobos_case_insensitive,
    mode=mode, lmodes=lmodes, print_fobos_template=print_fobos_template,
    only_obj=only_obj, only_target=only_target,
    compiler=compiler.lower(), fc=fc, cflags=cflags, lflags=lflags, modsw=modsw,
    mpi=mpi, openmp=openmp, openmp_offload=openmp_offload,
    coarray=coarray, coverage=coverage, profile=profile,
    mklib=mklib, ar=ar, arflags=arflags, ranlib=ranlib,
    cflags_heritage=cflags_heritage, track_build=track_build,
    src=src or ['./'],
    build_dir=build_dir, obj_dir=obj_dir, mod_dir=mod_dir,
    lib_dir=lib_dir or [], include=include or [],
    exclude_dirs=exclude_dirs or [],
    disable_recursive_search=disable_recursive_search,
    target=target, output=output,
    exclude=exclude or [], libs=libs or [], vlibs=vlibs or [],
    ext_libs=ext_libs or [], ext_vlibs=ext_vlibs or [],
    dependon=dependon or [],
    inc=inc or list(__extensions_inc__),
    extensions=extensions or list(__extensions_parsed__),
    build_all=build_all,
    force_compile=force_compile, colors=colors, log=log, graph=graph,
    quiet=quiet, verbose=verbose, jobs=jobs, makefile=makefile,
  )

# ---------------------------------------------------------------------------
# rule command
# ---------------------------------------------------------------------------

@app.command('rule')
def cmd_rule(
  ctx: typer.Context,
  # fobos group
  fobos:                    Annotated[Optional[str],       typer.Option('--fobos', '-f',           help='Specify a "fobos" file named differently from "fobos"')] = None,
  fobos_case_insensitive:   Annotated[bool,                typer.Option('--fci',                   help='Assume fobos inputs as case insensitive')] = False,
  mode:                     Annotated[Optional[str],       typer.Option('--mode',                  help='Select a mode defined into a fobos file', autocompletion=_complete_fobos_mode)] = None,
  lmodes:                   Annotated[bool,                typer.Option('--lmodes',                help='List the modes defined into a fobos file')] = False,
  print_fobos_template:     Annotated[bool,                typer.Option('--print-fobos-template',  help='Print a comprehensive fobos template')] = False,
  # compiler group
  compiler:                 Annotated[str,                 typer.Option('--compiler',              help='Compiler used (case insensitive, default gnu)', autocompletion=_complete_compiler)] = 'gnu',
  fc:                       Annotated[Optional[str],       typer.Option('--fc',                    help='Fortran compiler statement (for --compiler custom)')] = None,
  cflags:                   Annotated[Optional[str],       typer.Option('--cflags',                help='Compile flags')] = None,
  lflags:                   Annotated[Optional[str],       typer.Option('--lflags',                help='Link flags')] = None,
  modsw:                    Annotated[Optional[str],       typer.Option('--modsw',                 help='Module search path switch (for --compiler custom)')] = None,
  mpi:                      Annotated[bool,                typer.Option('--mpi',                   help='Use MPI enabled version of compiler')] = False,
  openmp:                   Annotated[bool,                typer.Option('--openmp',                help='Use OpenMP pragmas')] = False,
  openmp_offload:           Annotated[bool,                typer.Option('--openmp-offload',        help='Use OpenMP Offload pragmas')] = False,
  coarray:                  Annotated[bool,                typer.Option('--coarray',               help='Use coarrays')] = False,
  coverage:                 Annotated[bool,                typer.Option('--coverage',              help='Instrument for coverage analysis')] = False,
  profile:                  Annotated[bool,                typer.Option('--profile',               help='Instrument for profiling analysis')] = False,
  mklib:                    Annotated[Optional[str],       typer.Option('--mklib',                 help='Build library instead of program (static|shared)', autocompletion=_complete_mklib)] = None,
  ar:                       Annotated[str,                 typer.Option('--ar',                    help='Archiver executable for static libraries [default: ar]')] = 'ar',
  arflags:                  Annotated[str,                 typer.Option('--arflags',               help='Archiver flags for static libraries [default: -rcs]')] = '-rcs',
  ranlib:                   Annotated[str,                 typer.Option('--ranlib',                help='Ranlib executable; empty string to skip [default: ranlib]')] = 'ranlib',
  cflags_heritage:          Annotated[bool,                typer.Option('--cflags-heritage', '--ch', help='Store cflags for heritage checking; re-compile all if they change')] = False,
  track_build:              Annotated[bool,                typer.Option('--track-build', '--tb',   help='Store build info for the install command')] = False,
  # directories group
  src:                      Annotated[Optional[List[str]], typer.Option('--src', '-s',             help='Root-directory of source files [default: ./]')] = None,
  build_dir:                Annotated[str,                 typer.Option('--build-dir', '--dbld',   help='Directory containing built objects [default: ./]')] = './',
  obj_dir:                  Annotated[str,                 typer.Option('--obj-dir', '--dobj',     help='Directory containing compiled objects [default: ./obj/]')] = './obj/',
  mod_dir:                  Annotated[str,                 typer.Option('--mod-dir', '--dmod',     help='Directory containing .mod files [default: ./mod/]')] = './mod/',
  lib_dir:                  Annotated[Optional[List[str]], typer.Option('--lib-dir', '--dlib',     help='Directories searched for libraries')] = None,
  include:                  Annotated[Optional[List[str]], typer.Option('--include', '-i',         help='Directories for searching included files')] = None,
  exclude_dirs:             Annotated[Optional[List[str]], typer.Option('--exclude-dirs', '--ed',  help='Exclude directories from the building process')] = None,
  disable_recursive_search: Annotated[bool,                typer.Option('--disable-recursive-search', '--drs', help='Disable recursive search inside directories')] = False,
  # files group
  target:                   Annotated[Optional[str],       typer.Option('--target', '-t',          help='Specify a target file')] = None,
  output:                   Annotated[Optional[str],       typer.Option('--output', '-o',          help='Output file name (used with --target)')] = None,
  exclude:                  Annotated[Optional[List[str]], typer.Option('--exclude', '-e',         help='Exclude files from the building process')] = None,
  libs:                     Annotated[Optional[List[str]], typer.Option('--libs',                  help='External libraries with full paths')] = None,
  vlibs:                    Annotated[Optional[List[str]], typer.Option('--vlibs',                 help='Volatile external libraries with full paths')] = None,
  ext_libs:                 Annotated[Optional[List[str]], typer.Option('--ext-libs',              help='External libraries in compiler path')] = None,
  ext_vlibs:                Annotated[Optional[List[str]], typer.Option('--ext-vlibs',             help='Volatile external libraries in compiler path')] = None,
  dependon:                 Annotated[Optional[List[str]], typer.Option('--dependon',              help='Interdependent external fobos files for interdependent building')] = None,
  inc:                      Annotated[Optional[List[str]], typer.Option('--inc',                   help='Extensions for include files', autocompletion=_complete_extensions)] = None,
  extensions:               Annotated[Optional[List[str]], typer.Option('--extensions',            help='Extensions of parsed files', autocompletion=_complete_extensions)] = None,
  build_all:                Annotated[bool,                typer.Option('--build-all',             help='Build all sources parsed')] = False,
  # preprocessor group
  preprocessor:             Annotated[Optional[str],       typer.Option('--preprocessor',          help='Preprocessor name (e.g. PreForM.py)')] = None,
  preproc:                  Annotated[Optional[str],       typer.Option('--preproc',               help='Preprocessor flags for the main compiler')] = None,
  preprocessor_args:        Annotated[str,                 typer.Option('--preprocessor-args', '--app', help='Preprocessor-specific flags')] = '',
  preprocessor_no_o:        Annotated[bool,                typer.Option('--preprocessor-no-o', '--npp', help='Do not add -o in front of output in preprocessor command')] = False,
  preprocessor_dir:         Annotated[Optional[str],       typer.Option('--preprocessor-dir', '--dpp', help='Directory for preprocessed files')] = None,
  preprocessor_ext:         Annotated[Optional[List[str]], typer.Option('--preprocessor-ext', '--epp', help='File extensions to preprocess')] = None,
  # rules group
  execute:                  Annotated[Optional[str],       typer.Option('--execute', '--ex',       help='Specify a rule (defined in fobos) to execute')] = None,
  list_rules:               Annotated[bool,                typer.Option('--list', '--ls',          help='List the rules defined in a fobos file')] = False,
  # intrinsic rules group
  get:                      Annotated[Optional[str],       typer.Option('--get',                   help='Get option value defined in fobos (e.g. --get build_dir)')] = None,
  get_output_name:          Annotated[bool,                typer.Option('--get-output-name',       help='Get the final output name from fobos options')] = False,
  ford:                     Annotated[Optional[str],       typer.Option('--ford',                  help='Build docs with Ford tool (specify project-file.md)')] = None,
  gcov_analyzer:            Annotated[Optional[List[str]], typer.Option('--gcov-analyzer',         help='Analyse .gcov files; args: REPORTS_DIR [SUMMARY_FILE]')] = None,
  is_ascii_kind_supported:  Annotated[bool,                typer.Option('--is-ascii-kind-supported',  help='Check if compiler supports ASCII kind')] = False,
  is_ucs4_kind_supported:   Annotated[bool,                typer.Option('--is-ucs4-kind-supported',   help='Check if compiler supports UCS4 kind')] = False,
  is_float128_kind_supported: Annotated[bool,              typer.Option('--is-float128-kind-supported', help='Check if compiler supports float128 kind')] = False,
  # fancy group
  force_compile:            Annotated[bool,                typer.Option('--force-compile',         help='Force to (re-)compile all')] = False,
  colors:                   Annotated[bool,                typer.Option('--colors',                help='Activate colors in shell prints')] = False,
  log:                      Annotated[bool,                typer.Option('--log', '-l',             help='Activate log file creation')] = False,
  graph:                    Annotated[bool,                typer.Option('--graph',                 help='Generate a dependencies graph via graphviz')] = False,
  quiet:                    Annotated[bool,                typer.Option('--quiet', '-q',           help='Less verbose output')] = False,
  verbose:                  Annotated[bool,                typer.Option('--verbose',               help='Extremely verbose output for debugging')] = False,
  jobs:                     Annotated[int,                 typer.Option('--jobs', '-j',            help='Number of concurrent compilation jobs [default: 1]')] = 1,
  makefile:                 Annotated[Optional[str],       typer.Option('--makefile', '-m',        help='Generate a GNU Makefile for building the project')] = None,
):
  """Execute special rules or user-defined ones from a fobos file."""
  ctx.ensure_object(dict)
  ctx.obj['cliargs'] = _ns(
    which='rule',
    fobos=fobos, fobos_case_insensitive=fobos_case_insensitive,
    mode=mode, lmodes=lmodes, print_fobos_template=print_fobos_template,
    compiler=compiler.lower(), fc=fc, cflags=cflags, lflags=lflags, modsw=modsw,
    mpi=mpi, openmp=openmp, openmp_offload=openmp_offload,
    coarray=coarray, coverage=coverage, profile=profile,
    mklib=mklib, ar=ar, arflags=arflags, ranlib=ranlib,
    cflags_heritage=cflags_heritage, track_build=track_build,
    src=src or ['./'],
    build_dir=build_dir, obj_dir=obj_dir, mod_dir=mod_dir,
    lib_dir=lib_dir or [], include=include or [],
    exclude_dirs=exclude_dirs or [],
    disable_recursive_search=disable_recursive_search,
    target=target, output=output,
    exclude=exclude or [], libs=libs or [], vlibs=vlibs or [],
    ext_libs=ext_libs or [], ext_vlibs=ext_vlibs or [],
    dependon=dependon or [],
    inc=inc or list(__extensions_inc__),
    extensions=extensions or list(__extensions_parsed__),
    build_all=build_all,
    preprocessor=preprocessor, preproc=preproc,
    preprocessor_args=preprocessor_args,
    preprocessor_no_o=preprocessor_no_o,
    preprocessor_dir=preprocessor_dir,
    preprocessor_ext=preprocessor_ext or [],
    execute=execute, list=list_rules,
    get=get, get_output_name=get_output_name,
    ford=ford, gcov_analyzer=gcov_analyzer,
    is_ascii_kind_supported=is_ascii_kind_supported,
    is_ucs4_kind_supported=is_ucs4_kind_supported,
    is_float128_kind_supported=is_float128_kind_supported,
    force_compile=force_compile, colors=colors, log=log, graph=graph,
    quiet=quiet, verbose=verbose, jobs=jobs, makefile=makefile,
  )

# ---------------------------------------------------------------------------
# install command
# ---------------------------------------------------------------------------

@app.command('install')
def cmd_install(
  ctx: typer.Context,
  repo:                     Annotated[Optional[str],       typer.Argument(               help='GitHub repository: "user/repo" shorthand or full HTTPS URL')] = None,
  # fobos group
  fobos:                    Annotated[Optional[str],       typer.Option('--fobos', '-f', help='Specify a "fobos" file named differently from "fobos"')] = None,
  fobos_case_insensitive:   Annotated[bool,                typer.Option('--fci',         help='Assume fobos inputs as case insensitive')] = False,
  mode:                     Annotated[Optional[str],       typer.Option('--mode',        help='Select a mode defined into a fobos file', autocompletion=_complete_fobos_mode)] = None,
  lmodes:                   Annotated[bool,                typer.Option('--lmodes',      help='List the modes defined into a fobos file')] = False,
  print_fobos_template:     Annotated[bool,                typer.Option('--print-fobos-template', help='Print a comprehensive fobos template')] = False,
  # directories group (install variant)
  build_dir:                Annotated[str,                 typer.Option('--build-dir', '--dbld', help='Directory containing built objects [default: ./]')] = './',
  prefix:                   Annotated[str,                 typer.Option('--prefix', '-p', help='Prefix path where built objects are installed [default: ./]')] = './',
  bin:                      Annotated[str,                 typer.Option('--bin',         help='Sub-directory for executables [default: bin/]')] = 'bin/',
  lib:                      Annotated[str,                 typer.Option('--lib',         help='Sub-directory for libraries [default: lib/]')] = 'lib/',
  include:                  Annotated[str,                 typer.Option('--include',     help='Sub-directory for include files [default: include/]')] = 'include/',
  # GitHub install options
  branch:                   Annotated[Optional[str],       typer.Option('--branch',      help='Branch to check out when installing from GitHub')] = None,
  tag:                      Annotated[Optional[str],       typer.Option('--tag',         help='Tag to check out when installing from GitHub')] = None,
  rev:                      Annotated[Optional[str],       typer.Option('--rev',         help='Commit revision to check out when installing from GitHub')] = None,
  update:                   Annotated[bool,                typer.Option('--update',      help='Re-fetch (git pull) before building and installing')] = False,
  no_build:                 Annotated[bool,                typer.Option('--no-build',    help='Clone only — skip building and installing')] = False,
  deps_dir:                 Annotated[Optional[str],       typer.Option('--deps-dir',    help='Directory for cloning GitHub repositories [default: ~/.fobis/]')] = None,
  # fancy group
  force_compile:            Annotated[bool,                typer.Option('--force-compile', help='Force to (re-)compile all')] = False,
  colors:                   Annotated[bool,                typer.Option('--colors',      help='Activate colors in shell prints')] = False,
  log:                      Annotated[bool,                typer.Option('--log', '-l',   help='Activate log file creation')] = False,
  graph:                    Annotated[bool,                typer.Option('--graph',       help='Generate a dependencies graph via graphviz')] = False,
  quiet:                    Annotated[bool,                typer.Option('--quiet', '-q', help='Less verbose output')] = False,
  verbose:                  Annotated[bool,                typer.Option('--verbose',     help='Extremely verbose output for debugging')] = False,
  jobs:                     Annotated[int,                 typer.Option('--jobs', '-j',  help='Number of concurrent compilation jobs [default: 1]')] = 1,
  makefile:                 Annotated[Optional[str],       typer.Option('--makefile', '-m', help='Generate a GNU Makefile for building the project')] = None,
):
  """Install previously built files, or clone+build+install a GitHub-hosted FoBiS project."""
  ctx.ensure_object(dict)
  ctx.obj['cliargs'] = _ns(
    which='install',
    repo=repo,
    fobos=fobos, fobos_case_insensitive=fobos_case_insensitive,
    mode=mode, lmodes=lmodes, print_fobos_template=print_fobos_template,
    build_dir=build_dir, prefix=prefix,
    bin=bin, lib=lib, include=include,
    branch=branch, tag=tag, rev=rev,
    update=update, no_build=no_build, deps_dir=deps_dir,
    force_compile=force_compile, colors=colors, log=log, graph=graph,
    quiet=quiet, verbose=verbose, jobs=jobs, makefile=makefile,
  )

# ---------------------------------------------------------------------------
# doctests command
# ---------------------------------------------------------------------------

@app.command('doctests')
def cmd_doctests(
  ctx: typer.Context,
  # fobos group
  fobos:                    Annotated[Optional[str],       typer.Option('--fobos', '-f',           help='Specify a "fobos" file named differently from "fobos"')] = None,
  fobos_case_insensitive:   Annotated[bool,                typer.Option('--fci',                   help='Assume fobos inputs as case insensitive')] = False,
  mode:                     Annotated[Optional[str],       typer.Option('--mode',                  help='Select a mode defined into a fobos file', autocompletion=_complete_fobos_mode)] = None,
  lmodes:                   Annotated[bool,                typer.Option('--lmodes',                help='List the modes defined into a fobos file')] = False,
  print_fobos_template:     Annotated[bool,                typer.Option('--print-fobos-template',  help='Print a comprehensive fobos template')] = False,
  # compiler group
  compiler:                 Annotated[str,                 typer.Option('--compiler',              help='Compiler used (case insensitive, default gnu)', autocompletion=_complete_compiler)] = 'gnu',
  fc:                       Annotated[Optional[str],       typer.Option('--fc',                    help='Fortran compiler statement (for --compiler custom)')] = None,
  cflags:                   Annotated[Optional[str],       typer.Option('--cflags',                help='Compile flags')] = None,
  lflags:                   Annotated[Optional[str],       typer.Option('--lflags',                help='Link flags')] = None,
  modsw:                    Annotated[Optional[str],       typer.Option('--modsw',                 help='Module search path switch (for --compiler custom)')] = None,
  mpi:                      Annotated[bool,                typer.Option('--mpi',                   help='Use MPI enabled version of compiler')] = False,
  openmp:                   Annotated[bool,                typer.Option('--openmp',                help='Use OpenMP pragmas')] = False,
  openmp_offload:           Annotated[bool,                typer.Option('--openmp-offload',        help='Use OpenMP Offload pragmas')] = False,
  coarray:                  Annotated[bool,                typer.Option('--coarray',               help='Use coarrays')] = False,
  coverage:                 Annotated[bool,                typer.Option('--coverage',              help='Instrument for coverage analysis')] = False,
  profile:                  Annotated[bool,                typer.Option('--profile',               help='Instrument for profiling analysis')] = False,
  mklib:                    Annotated[Optional[str],       typer.Option('--mklib',                 help='Build library instead of program (static|shared)', autocompletion=_complete_mklib)] = None,
  ar:                       Annotated[str,                 typer.Option('--ar',                    help='Archiver executable for static libraries [default: ar]')] = 'ar',
  arflags:                  Annotated[str,                 typer.Option('--arflags',               help='Archiver flags for static libraries [default: -rcs]')] = '-rcs',
  ranlib:                   Annotated[str,                 typer.Option('--ranlib',                help='Ranlib executable; empty string to skip [default: ranlib]')] = 'ranlib',
  cflags_heritage:          Annotated[bool,                typer.Option('--cflags-heritage', '--ch', help='Store cflags for heritage checking')] = False,
  track_build:              Annotated[bool,                typer.Option('--track-build', '--tb',   help='Store build info for the install command')] = False,
  # directories group
  src:                      Annotated[Optional[List[str]], typer.Option('--src', '-s',             help='Root-directory of source files [default: ./]')] = None,
  build_dir:                Annotated[str,                 typer.Option('--build-dir', '--dbld',   help='Directory containing built objects [default: ./]')] = './',
  obj_dir:                  Annotated[str,                 typer.Option('--obj-dir', '--dobj',     help='Directory containing compiled objects [default: ./obj/]')] = './obj/',
  mod_dir:                  Annotated[str,                 typer.Option('--mod-dir', '--dmod',     help='Directory containing .mod files [default: ./mod/]')] = './mod/',
  lib_dir:                  Annotated[Optional[List[str]], typer.Option('--lib-dir', '--dlib',     help='Directories searched for libraries')] = None,
  include:                  Annotated[Optional[List[str]], typer.Option('--include', '-i',         help='Directories for searching included files')] = None,
  exclude_dirs:             Annotated[Optional[List[str]], typer.Option('--exclude-dirs', '--ed',  help='Exclude directories from the building process')] = None,
  disable_recursive_search: Annotated[bool,                typer.Option('--disable-recursive-search', '--drs', help='Disable recursive search inside directories')] = False,
  # files group (doctests variant)
  target:                   Annotated[Optional[str],       typer.Option('--target', '-t',          help='Specify a target file')] = None,
  output:                   Annotated[Optional[str],       typer.Option('--output', '-o',          help='Output file name (used with --target)')] = None,
  exclude:                  Annotated[Optional[List[str]], typer.Option('--exclude', '-e',         help='Exclude files from the building process')] = None,
  libs:                     Annotated[Optional[List[str]], typer.Option('--libs',                  help='External libraries with full paths')] = None,
  vlibs:                    Annotated[Optional[List[str]], typer.Option('--vlibs',                 help='Volatile external libraries with full paths')] = None,
  ext_libs:                 Annotated[Optional[List[str]], typer.Option('--ext-libs',              help='External libraries in compiler path')] = None,
  ext_vlibs:                Annotated[Optional[List[str]], typer.Option('--ext-vlibs',             help='Volatile external libraries in compiler path')] = None,
  dependon:                 Annotated[Optional[List[str]], typer.Option('--dependon',              help='Interdependent external fobos files')] = None,
  inc:                      Annotated[Optional[List[str]], typer.Option('--inc',                   help='Extensions for include files', autocompletion=_complete_extensions)] = None,
  extensions:               Annotated[Optional[List[str]], typer.Option('--extensions',            help='Extensions of parsed files', autocompletion=_complete_extensions)] = None,
  build_all:                Annotated[bool,                typer.Option('--build-all',             help='Build all sources parsed')] = False,
  keep_volatile_doctests:   Annotated[bool,                typer.Option('--keep-volatile-doctests', help='Keep the volatile doctests programs')] = False,
  exclude_from_doctests:    Annotated[Optional[List[str]], typer.Option('--exclude-from-doctests', help='Exclude files from the doctests process')] = None,
  # preprocessor group (doctests variant)
  preprocessor:             Annotated[Optional[str],       typer.Option('--preprocessor',          help='Preprocessor name (e.g. PreForM.py)')] = None,
  preproc:                  Annotated[Optional[str],       typer.Option('--preproc',               help='Preprocessor flags for the main compiler')] = None,
  preprocessor_args:        Annotated[str,                 typer.Option('--preprocessor-args', '--app', help='Preprocessor-specific flags')] = '',
  preprocessor_no_o:        Annotated[bool,                typer.Option('--preprocessor-no-o', '--npp', help='Do not add -o in front of output in preprocessor command')] = False,
  preprocessor_dir:         Annotated[Optional[str],       typer.Option('--preprocessor-dir', '--dpp', help='Directory for preprocessed files')] = None,
  preprocessor_ext:         Annotated[Optional[List[str]], typer.Option('--preprocessor-ext', '--epp', help='File extensions to preprocess')] = None,
  doctests_preprocessor:    Annotated[str,                 typer.Option('--doctests-preprocessor', help='Preprocessor used during doctests parsing (cpp|fpp)', autocompletion=_complete_doctests_preprocessor)] = 'cpp',
  # fancy group
  force_compile:            Annotated[bool,                typer.Option('--force-compile',         help='Force to (re-)compile all')] = False,
  colors:                   Annotated[bool,                typer.Option('--colors',                help='Activate colors in shell prints')] = False,
  log:                      Annotated[bool,                typer.Option('--log', '-l',             help='Activate log file creation')] = False,
  graph:                    Annotated[bool,                typer.Option('--graph',                 help='Generate a dependencies graph via graphviz')] = False,
  quiet:                    Annotated[bool,                typer.Option('--quiet', '-q',           help='Less verbose output')] = False,
  verbose:                  Annotated[bool,                typer.Option('--verbose',               help='Extremely verbose output for debugging')] = False,
  jobs:                     Annotated[int,                 typer.Option('--jobs', '-j',            help='Number of concurrent compilation jobs [default: 1]')] = 1,
  makefile:                 Annotated[Optional[str],       typer.Option('--makefile', '-m',        help='Generate a GNU Makefile for building the project')] = None,
):
  """Test all valid doctests snippets found."""
  ctx.ensure_object(dict)
  ctx.obj['cliargs'] = _ns(
    which='doctests',
    fobos=fobos, fobos_case_insensitive=fobos_case_insensitive,
    mode=mode, lmodes=lmodes, print_fobos_template=print_fobos_template,
    compiler=compiler.lower(), fc=fc, cflags=cflags, lflags=lflags, modsw=modsw,
    mpi=mpi, openmp=openmp, openmp_offload=openmp_offload,
    coarray=coarray, coverage=coverage, profile=profile,
    mklib=mklib, ar=ar, arflags=arflags, ranlib=ranlib,
    cflags_heritage=cflags_heritage, track_build=track_build,
    src=src or ['./'],
    build_dir=build_dir, obj_dir=obj_dir, mod_dir=mod_dir,
    lib_dir=lib_dir or [], include=include or [],
    exclude_dirs=exclude_dirs or [],
    disable_recursive_search=disable_recursive_search,
    target=target, output=output,
    exclude=exclude or [], libs=libs or [], vlibs=vlibs or [],
    ext_libs=ext_libs or [], ext_vlibs=ext_vlibs or [],
    dependon=dependon or [],
    inc=inc or list(__extensions_inc__),
    extensions=extensions or list(__extensions_parsed__),
    build_all=build_all,
    keep_volatile_doctests=keep_volatile_doctests,
    exclude_from_doctests=exclude_from_doctests or [],
    preprocessor=preprocessor, preproc=preproc,
    preprocessor_args=preprocessor_args,
    preprocessor_no_o=preprocessor_no_o,
    preprocessor_dir=preprocessor_dir,
    preprocessor_ext=preprocessor_ext or [],
    doctests_preprocessor=doctests_preprocessor,
    force_compile=force_compile, colors=colors, log=log, graph=graph,
    quiet=quiet, verbose=verbose, jobs=jobs, makefile=makefile,
  )

# ---------------------------------------------------------------------------
# fetch command
# ---------------------------------------------------------------------------

@app.command('fetch')
def cmd_fetch(
  ctx: typer.Context,
  # fobos group
  fobos:                    Annotated[Optional[str],       typer.Option('--fobos', '-f', help='Specify a "fobos" file named differently from "fobos"')] = None,
  fobos_case_insensitive:   Annotated[bool,                typer.Option('--fci',         help='Assume fobos inputs as case insensitive')] = False,
  mode:                     Annotated[Optional[str],       typer.Option('--mode',        help='Select a mode defined into a fobos file', autocompletion=_complete_fobos_mode)] = None,
  lmodes:                   Annotated[bool,                typer.Option('--lmodes',      help='List the modes defined into a fobos file')] = False,
  print_fobos_template:     Annotated[bool,                typer.Option('--print-fobos-template', help='Print a comprehensive fobos template')] = False,
  # fetch-specific options
  deps_dir:                 Annotated[Optional[str],       typer.Option('--deps-dir',    help='Directory for storing fetched dependencies [default: .fobis_deps or fobos [dependencies] setting]')] = None,
  update:                   Annotated[bool,                typer.Option('--update',      help='Update already-fetched dependencies (git fetch + re-checkout)')] = False,
  no_build:                 Annotated[bool,                typer.Option('--no-build',    help='Only fetch dependencies, do not build them')] = False,
  # fancy group
  force_compile:            Annotated[bool,                typer.Option('--force-compile', help='Force to (re-)compile all')] = False,
  colors:                   Annotated[bool,                typer.Option('--colors',      help='Activate colors in shell prints')] = False,
  log:                      Annotated[bool,                typer.Option('--log', '-l',   help='Activate log file creation')] = False,
  graph:                    Annotated[bool,                typer.Option('--graph',       help='Generate a dependencies graph via graphviz')] = False,
  quiet:                    Annotated[bool,                typer.Option('--quiet', '-q', help='Less verbose output')] = False,
  verbose:                  Annotated[bool,                typer.Option('--verbose',     help='Extremely verbose output for debugging')] = False,
  jobs:                     Annotated[int,                 typer.Option('--jobs', '-j',  help='Number of concurrent compilation jobs [default: 1]')] = 1,
  makefile:                 Annotated[Optional[str],       typer.Option('--makefile', '-m', help='Generate a GNU Makefile for building the project')] = None,
):
  """Fetch and build GitHub-hosted Fortran dependencies listed in the fobos [dependencies] section."""
  ctx.ensure_object(dict)
  ctx.obj['cliargs'] = _ns(
    which='fetch',
    fobos=fobos, fobos_case_insensitive=fobos_case_insensitive,
    mode=mode, lmodes=lmodes, print_fobos_template=print_fobos_template,
    deps_dir=deps_dir, update=update, no_build=no_build,
    force_compile=force_compile, colors=colors, log=log, graph=graph,
    quiet=quiet, verbose=verbose, jobs=jobs, makefile=makefile,
  )



#!/usr/bin/env python
"""
FoBiS.py, Fortran Building System for poor men
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# modules loading
# from __future__ import print_function
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import unicode_literals
# from future import standard_library
# standard_library.install_aliases()
# from builtins import *
try:
  import configparser as configparser
except ImportError:
  import configparser
import os
import shutil
import sys
from .Builder import Builder
from .Cleaner import Cleaner
from .FoBiSConfig import FoBiSConfig
from .Gcov import Gcov, _mermaid_pie
from .ParsedFile import ParsedFile
from .utils import dependency_hiearchy
from .utils import remove_other_main
from .utils import syswork
from .utils import safe_mkdir


def main():
  """
  Main function.
  """
  run_fobis()
  sys.exit(0)


def run_fobis(fake_args=None):
  """
  Run FoBiS accordingly to the user configuration.

  Parameters
  ----------
  fake_args : list
    list containing fake CLAs for using without CLI
  """
  configuration = FoBiSConfig(fake_args=fake_args)
  if configuration.cliargs.which == 'rule':
    run_fobis_rule(configuration)
  else:
    if configuration.cliargs.lmodes:
      configuration.fobos.modes_list()
      sys.exit(0)
    if configuration.cliargs.print_fobos_template:
      configuration.fobos.print_template(configuration.cliargs)
      sys.exit(0)
    if configuration.cliargs.which == 'clean':
      run_fobis_clean(configuration)
    if configuration.cliargs.which == 'build':
      run_fobis_build(configuration)
    if configuration.cliargs.which == 'install':
      run_fobis_install(configuration)
    if configuration.cliargs.which == 'doctests':
      run_fobis_doctests(configuration)
    if configuration.cliargs.which == 'fetch':
      run_fobis_fetch(configuration)
  return


def run_fobis_fetch(configuration):
  """
  Run FoBiS in fetch mode: clone/update and optionally build GitHub dependencies.

  Parameters
  ----------
  configuration : FoBiSConfig()
  """
  from .Fetcher import Fetcher
  deps_dir = configuration.cliargs.deps_dir or configuration.fobos.get_deps_dir()
  fetcher = Fetcher(
    deps_dir=deps_dir,
    print_n=configuration.print_b,
    print_w=configuration.print_r)
  deps = configuration.fobos.get_dependencies()
  if not deps:
    configuration.print_r('No [dependencies] section found in fobos file')
    return
  deps_info = []
  for name, spec in deps.items():
    parsed = fetcher.parse_dep_spec(spec)
    use_mode = parsed.get('use', 'sources')
    dep_path = fetcher.fetch(
      name,
      parsed['url'],
      branch=parsed.get('branch'),
      tag=parsed.get('tag'),
      rev=parsed.get('rev'),
      update=configuration.cliargs.update)
    if not configuration.cliargs.no_build and use_mode == 'fobos':
      fetcher.build_dep(name, dep_path, mode=parsed.get('mode'))
    deps_info.append({'name': name, 'path': dep_path, 'mode': parsed.get('mode', ''), 'use': use_mode})
  fetcher.save_config(deps_info)


def run_fobis_build(configuration):
  """
  Run FoBiS in build mode.

  Parameters
  ----------
  configuration : FoBiSConfig()
  """

  builder = Builder(cliargs=configuration.cliargs, print_n=configuration.print_b, print_w=configuration.print_r)
  pfiles = parse_files(configuration=configuration)
  dependency_hiearchy(builder=builder, pfiles=pfiles, print_w=configuration.print_r, force_compile=configuration.cliargs.force_compile)
  if configuration.cliargs.makefile:
    save_makefile(configuration=configuration, pfiles=pfiles, builder=builder)
    return
  nomodlibs = build_nomodlibs(configuration=configuration, pfiles=pfiles, builder=builder)
  submodules = build_submodules(configuration=configuration, pfiles=pfiles, builder=builder)
  # building target or all programs found
  for pfile in pfiles:
    if configuration.cliargs.build_all:
      build_pfile(configuration=configuration, pfile=pfile, pfiles=pfiles, nomodlibs=nomodlibs, submodules=submodules, builder=builder)
    else:
      if configuration.cliargs.target:
        if os.path.basename(configuration.cliargs.target) == os.path.basename(pfile.name):
          build_pfile(configuration=configuration, pfile=pfile, pfiles=pfiles, nomodlibs=nomodlibs, submodules=submodules, builder=builder)
      else:
        if pfile.program:
          build_pfile(configuration=configuration, pfile=pfile, pfiles=pfiles, nomodlibs=nomodlibs, submodules=submodules, builder=builder)


def run_fobis_clean(configuration):
  """
  Run FoBiS in build mode.

  Parameters
  ----------
  configuration : FoBiSConfig()
  """
  cleaner = Cleaner(cliargs=configuration.cliargs, print_w=configuration.print_r)
  if not configuration.cliargs.only_obj and not configuration.cliargs.only_target:
    cleaner.clean_mod()
    cleaner.clean_obj()
    cleaner.clean_target()
  if configuration.cliargs.only_obj:
    cleaner.clean_mod()
    cleaner.clean_obj()
  if configuration.cliargs.only_target:
    cleaner.clean_target()


def run_fobis_rule(configuration):
  """
  Run FoBiS in build mode.

  Parameters
  ----------
  configuration : FoBiSConfig()
  """
  if configuration.cliargs.list:
    configuration.fobos.rules_list(quiet=configuration.cliargs.quiet)
  elif configuration.cliargs.execute:
    configuration.fobos.rule_execute(rule=configuration.cliargs.execute, quiet=configuration.cliargs.quiet)
  elif configuration.cliargs.get:
    configuration.fobos.get(option=configuration.cliargs.get, mode=configuration.cliargs.mode)
  elif configuration.cliargs.get_output_name:
    configuration.fobos.get_output_name(mode=configuration.cliargs.mode)
  elif configuration.cliargs.ford:
    result = syswork("ford " + configuration.cliargs.ford)
    if result[0] != 0:
      configuration.print_r(result[1])
    else:
      configuration.print_b(result[1])
  elif configuration.cliargs.gcov_analyzer:
    gcov_analyzer(configuration=configuration)
  elif configuration.cliargs.is_ascii_kind_supported:
    is_ascii_kind_supported(configuration=configuration)
  elif configuration.cliargs.is_ucs4_kind_supported:
    is_ucs4_kind_supported(configuration=configuration)
  elif configuration.cliargs.is_float128_kind_supported:
    is_float128_kind_supported(configuration=configuration)


def run_fobis_install(configuration):
  """
  Run FoBiS in install mode.

  When 'repo' is provided (e.g. 'szaghi/FLAP' or a full URL) the command
  clones the GitHub-hosted FoBiS project, builds it with --track_build, and
  installs the resulting artifacts to the specified prefix.

  Without 'repo', the existing behaviour is preserved: install previously
  built artifacts recorded in .track_build files inside the build directory.

  Parameters
  ----------
  configuration : FoBiSConfig()
  """
  if configuration.cliargs.repo:
    from .Fetcher import Fetcher
    deps_dir = configuration.cliargs.deps_dir or os.path.join(os.path.expanduser('~'), '.fobis')
    fetcher = Fetcher(
      deps_dir=deps_dir,
      print_n=configuration.print_b,
      print_w=configuration.print_r)
    fetcher.install_from_github(
      repo=configuration.cliargs.repo,
      branch=configuration.cliargs.branch,
      tag=configuration.cliargs.tag,
      rev=configuration.cliargs.rev,
      mode=configuration.cliargs.mode,
      update=configuration.cliargs.update,
      no_build=configuration.cliargs.no_build,
      prefix=configuration.cliargs.prefix,
      bin_dir=configuration.cliargs.bin,
      lib_dir=configuration.cliargs.lib,
      include_dir=configuration.cliargs.include)
    return

  if not os.path.exists(configuration.cliargs.build_dir):
    configuration.fobos.print_w('Error: build directory not found! Maybe you have to run "FoBiS.py build" before.')
    sys.exit(1)
  safe_mkdir(directory=configuration.cliargs.prefix)
  for filename in os.listdir(configuration.cliargs.build_dir):
    if filename.endswith('.track_build'):
      is_program = False
      is_library = False
      track_file = configparser.ConfigParser()
      track_file.read(os.path.join(configuration.cliargs.build_dir, filename))
      if track_file.has_option(section='build', option='output'):
        output = track_file.get(section='build', option='output')
        if track_file.has_option(section='build', option='program'):
          is_program = track_file.get(section='build', option='program')
        if track_file.has_option(section='build', option='library'):
          is_library = track_file.get(section='build', option='library')
        if is_program:
          bin_path = os.path.join(configuration.cliargs.prefix, configuration.cliargs.bin)
          safe_mkdir(directory=bin_path)
          configuration.fobos.print_n('Install "' + output + '" in "' + bin_path + '"')
          shutil.copy(output, bin_path)
        if is_library:
          lib_path = os.path.join(configuration.cliargs.prefix, configuration.cliargs.lib)
          safe_mkdir(directory=lib_path)
          configuration.fobos.print_n('Install "' + output + '" in "' + lib_path + '"')
          shutil.copy(output, lib_path)
          if track_file.has_option(section='build', option='mod_file'):
            mod_file = track_file.get(section='build', option='mod_file')
            inc_path = os.path.join(configuration.cliargs.prefix, configuration.cliargs.include)
            safe_mkdir(directory=inc_path)
            configuration.fobos.print_n('Install "' + mod_file + '" in "' + inc_path + '"')
            shutil.copy(mod_file, inc_path)


def run_fobis_doctests(configuration):
  """
  Run FoBiS in build mode.

  Parameters
  ----------
  configuration : FoBiSConfig()
  """
  builder = Builder(cliargs=configuration.cliargs, print_n=configuration.print_b, print_w=configuration.print_r)
  pfiles = parse_files(configuration=configuration)
  dependency_hiearchy(builder=builder, pfiles=pfiles, print_w=configuration.print_r, force_compile=configuration.cliargs.force_compile)
  nomodlibs = build_nomodlibs(configuration=configuration, pfiles=pfiles, builder=builder)
  doctests, doctests_dirs = parse_doctests(configuration=configuration, pfiles=pfiles, builder=builder)
  for pfile in pfiles:
    doctests.append(pfile)
  dependency_hiearchy(builder=builder, pfiles=doctests, print_w=configuration.print_r, force_compile=configuration.cliargs.force_compile)
  test_doctests(configuration=configuration, doctests=doctests, pfiles=pfiles, nomodlibs=nomodlibs, builder=builder)
  if not configuration.cliargs.keep_volatile_doctests:
    for doc_dir in doctests_dirs:
      if os.path.isdir(doc_dir):
        shutil.rmtree(doc_dir)


def parse_files(configuration, src_dir=None, is_doctest=False):
  """
  Parse files and return the list of parsed files.

  Parameters
  ----------
  configuration : FoBiSConfig()
  src_dir: str
    root directory into which search; if omitted use configuration.cliargs.src

  Returns
  -------
  list
    list of ParsedFile() objects
  """
  pfiles = []
  if src_dir is not None:
    src = [src_dir]
  else:
    src = configuration.cliargs.src
  for src_dir in src:
    if configuration.cliargs.disable_recursive_search:
      for filename in os.listdir(src_dir):
        if any(os.path.splitext(os.path.basename(filename))[1] == ext for ext in configuration.cliargs.extensions):
          if (os.path.basename(filename) not in [os.path.basename(exc) for exc in configuration.cliargs.exclude] and
              all(exc not in os.path.dirname(filename) for exc in configuration.cliargs.exclude_dirs)):
             pfile = ParsedFile(name=os.path.join(src_dir, filename), is_doctest=is_doctest)
             if is_doctest:
               pfile.parse(inc=configuration.cliargs.inc, preprocessor=configuration.cliargs.doctests_preprocessor, preproc=configuration.cliargs.preproc, include=configuration.cliargs.include)
             else:
               pfile.parse(inc=configuration.cliargs.inc, preproc=configuration.cliargs.preproc, include=configuration.cliargs.include)
             pfiles.append(pfile)
    else:
      for root, _, files in os.walk(src_dir):
        for filename in files:
          if any(os.path.splitext(os.path.basename(filename))[1] == ext for ext in configuration.cliargs.extensions):
            if (os.path.basename(filename) not in [os.path.basename(exc) for exc in configuration.cliargs.exclude] and
                all(exc not in root for exc in configuration.cliargs.exclude_dirs)):
              filen = os.path.join(root, filename)
              pfile = ParsedFile(name=filen, is_doctest=is_doctest)
              if is_doctest:
                pfile.parse(inc=configuration.cliargs.inc, preprocessor=configuration.cliargs.doctests_preprocessor, preproc=configuration.cliargs.preproc, include=configuration.cliargs.include)
              else:
                pfile.parse(inc=configuration.cliargs.inc, preproc=configuration.cliargs.preproc, include=configuration.cliargs.include)
              pfiles.append(pfile)
  return pfiles


def parse_doctests(configuration, pfiles, builder):
  """Parse parsed-files for

  Parameters
  ----------
  configuration : FoBiSConfig()
  pfiles : list
    list of ParsedFile() objects
  builder : Builder()

  Returns
  -------
  list
    list of doctests
  """
  doctests = []
  doctests_dirs = []
  for pfile in pfiles:
    if pfile.doctest:
      if pfile.doctest.to_test:
        doc_dir = pfile.doctest.save_volatile_programs(build_dir=builder.build_dir)
        doctests_dirs.append(doc_dir)
  if len(doctests_dirs) > 0:
    doctests_dirs = list(set(doctests_dirs))
  for doc_dir in doctests_dirs:
    doctests += parse_files(configuration=configuration, src_dir=doc_dir, is_doctest=True)
  return doctests, doctests_dirs


def build_pfile(configuration, pfile, pfiles, nomodlibs, submodules, builder):
  """Build a parsed file.

  Parameters
  ----------
  configuration : FoBiSConfig()
  pfile : ParsedFile()
  pfiles : list
    list of ParsedFile() objects
  nomodlibs : list
    list of built non module libraries object names
  submodules : list
    list of built submodules object names
  builder : Builder()
  """
  configuration.print_b(builder.verbose(quiet=configuration.cliargs.quiet))
  if pfile.program:
    remove_other_main(builder=builder, pfiles=pfiles, myself=pfile)
  builder.build(file_to_build=pfile, output=configuration.cliargs.output, nomodlibs=nomodlibs, submodules=submodules, mklib=configuration.cliargs.mklib, verbose=configuration.cliargs.verbose, log=configuration.cliargs.log, track=configuration.cliargs.track_build)
  if configuration.cliargs.log:
    pfile.save_build_log(builder=builder)
  if configuration.cliargs.graph:
    pfile.save_dep_graph()


def build_nomodlibs(configuration, pfiles, builder):
  """Build all non module library files.

  Parameters
  ----------
  configuration : FoBiSConfig()
  pfiles : list
    list of ParsedFile() objects
  builder : Builder()

  Returns
  -------
  list
    list of built non module libraries object names
  """
  nomodlibs = []
  for pfile in pfiles:
    if pfile.nomodlib:
      build_ok = builder.build(file_to_build=pfile, verbose=configuration.cliargs.verbose, log=configuration.cliargs.log)
      if build_ok:
        nomodlibs = nomodlibs + pfile.obj_dependencies(exclude_programs=True)
  return nomodlibs


def build_submodules(configuration, pfiles, builder):
  """Build all submodule files.

  Parameters
  ----------
  configuration : FoBiSConfig()
  pfiles : list
    list of ParsedFile() objects
  builder : Builder()

  Returns
  -------
  list
    list of built submodules object names
  """
  submodules = []
  for pfile in pfiles:
    if pfile.submodule:
      build_ok = builder.build(file_to_build=pfile, verbose=configuration.cliargs.verbose, log=configuration.cliargs.log)
      if build_ok:
        submodules.append(pfile.basename + ".o")
        # Also include transitive module dependencies of this submodule.
        # Modules used exclusively by submodule files are not reachable from the
        # main program's module dependency graph and would otherwise be compiled
        # but never added to the link command.
        submodules += pfile.obj_dependencies(exclude_programs=True)
  return list(set(submodules))


def test_doctests(configuration, doctests, pfiles, nomodlibs, builder):
  """Test doctests: build/execute/check-result of each doctest.

  Parameters
  ----------
  configuration : FoBiSConfig()
  doctests : list
    list of ParsedFile() objects containing doctests
  pfiles : list
    list of ParsedFile() objects
  nomodlibs : list
    list of built non module libraries object names
  builder : Builder()
  """
  for test in doctests:
    if test.is_doctest and os.path.basename(test.name).split("-doctest")[0] not in [os.path.basename(os.path.splitext(exc)[0]) for exc in configuration.cliargs.exclude_from_doctests]:
      remove_other_main(builder=builder, pfiles=pfiles, myself=test)
      builder.build(file_to_build=test, nomodlibs=nomodlibs, verbose=False, log=False, quiet=True)
      test_exe = os.path.join(builder.build_dir, os.path.splitext(os.path.basename(test.name))[0])
      configuration.print_b('executing doctest ' + os.path.basename(test_exe))
      result = syswork(test_exe)
      if result[0] == 0:
        # comparing results
        test_result = os.path.join(os.path.dirname(test.name), os.path.splitext(os.path.basename(test.name))[0] + '.result')
        with open(test_result, 'r') as res:
          expected_result = res.read()
        if result[1].strip() == expected_result:
          configuration.print_b('doctest passed')
        else:
          configuration.print_r('doctest failed!')
          configuration.print_b('  result obtained: "' + result[1].strip() + '"')
          configuration.print_b('  result expected: "' + expected_result + '"')
      if not configuration.cliargs.keep_volatile_doctests:
        os.remove(test_exe)


def save_makefile(configuration, pfiles, builder):
  """
  Save GNU makefile.

  Parameters
  ----------
  pfiles : list
    list of parsed files
  builder : Builder object
  """
  def _gnu_variables(builder):
    """
    Method for getting GNU Make variables

    Parameters
    ----------
    builder : Builder object

    Returns
    -------
    str
      string containing the GNU Make variables
    """
    string = []
    string.append("\n#main building variables\n")
    string.append("DSRC    = " + " ".join(configuration.cliargs.src) + "\n")
    string.append(builder.gnu_make())
    string.append("VPATH   = $(DSRC) $(DOBJ) $(DMOD)" + "\n")
    string.append("MKDIRS  = $(DOBJ) $(DMOD) $(DEXE)" + "\n")
    string.append("LCEXES  = $(shell echo $(EXES) | tr '[:upper:]' '[:lower:]')\n")
    string.append("EXESPO  = $(addsuffix .o,$(LCEXES))\n")
    string.append("EXESOBJ = $(addprefix $(DOBJ),$(EXESPO))\n")
    string.append("\n")
    string.append("#auxiliary variables\n")
    string.append('COTEXT  = "Compiling $(<F)"\n')
    string.append('LITEXT  = "Assembling $@"\n')
    return "".join(string)

  def _gnu_building_rules(pfiles):
    """
    Method returing the building rules.

    Parameters
    ----------
    pfifles : list
      list of parsed file

    Returns
    -------
    str
      string containing the building rules
    """
    # collect non-module-libraries object
    nomodlibs = []
    for pfile in pfiles:
      if pfile.nomodlib:
        nomodlibs.append("$(DOBJ)" + pfile.basename.lower() + ".o")
    string = []
    string.append("\n#building rules\n")
    # linking rules
    for pfile in pfiles:
      save_target_rule = False
      if pfile.program:
        save_target_rule = True
      elif configuration.cliargs.target:
        if os.path.basename(configuration.cliargs.target) == os.path.basename(pfile.name):
          save_target_rule = True
      if save_target_rule:
        if len(nomodlibs) > 0:
          string.append("$(DEXE)" + pfile.basename.upper() + ": $(MKDIRS) " + "$(DOBJ)" + pfile.basename.lower() + ".o \\" + "\n")
          for nomod in nomodlibs[:-1]:
            string.append("\t" + nomod + " \\" + "\n")
          string.append("\t" + nomodlibs[-1] + "\n")
        else:
          string.append("$(DEXE)" + pfile.basename.upper() + ": $(MKDIRS) " + "$(DOBJ)" + pfile.basename.lower() + ".o\n")
        string.append("\t@rm -f $(filter-out $(DOBJ)" + pfile.basename.lower() + ".o,$(EXESOBJ))\n")
        string.append("\t@echo $(LITEXT)\n")
        string.append("\t@$(FC) $(OPTSL) $(DOBJ)*.o $(LIBS) -o $@\n")
        string.append("EXES := $(EXES) " + pfile.basename.upper() + "\n")
    return "".join(string)

  def _gnu_auxiliary_rules():
    """
    Method returing some useful GNU Make auxiliary rules

    Returns
    -------
    str
      string containing the auxiliary rules
    """
    string = []
    string.append("#phony auxiliary rules\n")
    string.append(".PHONY : $(MKDIRS)\n")
    string.append("$(MKDIRS):\n")
    string.append("\t@mkdir -p $@\n")
    string.append(".PHONY : cleanobj\n")
    string.append("cleanobj:\n")
    string.append("\t@echo deleting objects\n")
    string.append("\t@rm -fr $(DOBJ)\n")
    string.append(".PHONY : cleanmod\n")
    string.append("cleanmod:\n")
    string.append("\t@echo deleting mods\n")
    string.append("\t@rm -fr $(DMOD)\n")
    string.append(".PHONY : cleanexe\n")
    string.append("cleanexe:\n")
    string.append("\t@echo deleting exes\n")
    string.append("\t@rm -f $(addprefix $(DEXE),$(EXES))\n")
    string.append(".PHONY : clean\n")
    string.append("clean: cleanobj cleanmod\n")
    string.append(".PHONY : cleanall\n")
    string.append("cleanall: clean cleanexe\n")
    return "".join(string)

  string = []
  string.append("#!/usr/bin/make\n")
  string.append(_gnu_variables(builder=builder))
  string.append(_gnu_building_rules(pfiles=pfiles))
  string.append("\n#compiling rules\n")
  for pfile in pfiles:
    string.append(pfile.gnu_make_rule(builder=builder))
  string.append(_gnu_auxiliary_rules())
  with open(configuration.cliargs.makefile, "w") as mk_file:
    mk_file.writelines(string)


def gcov_analyzer(configuration):
  """
  Run gcov file analyzer.
  """
  gcovs = []
  for root, _, files in os.walk('.'):
    for filename in files:
      if os.path.splitext(os.path.basename(filename))[1] == '.gcov':
        configuration.print_b('Analyzing ' + os.path.join(root, filename))
        gcovs.append(Gcov(filename=os.path.join(root, filename)))
  for gcov in gcovs:
    gcov.parse()
    gcov.save(output=os.path.join(configuration.cliargs.gcov_analyzer[0], os.path.basename(gcov.filename) + '.md'), graphs=True)
  if len(gcovs) > 0 and len(configuration.cliargs.gcov_analyzer) > 1:
    string = []
    string.append('### ' + configuration.cliargs.gcov_analyzer[1] + '\n')
    for gcov in gcovs:
      string.append('\n#### [[' + os.path.basename(gcov.filename) + ']]\n\n')
      if gcov.l_pie_url or gcov.p_pie_url:
        string.append(gcov.l_pie_url + '\n' + gcov.p_pie_url + '\n')
      if gcov.metrics['coverage']:
        string.append('|Lines| | |\n')
        string.append('| --- | --- | --- |\n')
        string.append('|Executable lines            |' + gcov.metrics['coverage'][0] + '| |\n')
        string.append('|Executed lines              |' + gcov.metrics['coverage'][1] + '|' + gcov.metrics['coverage'][3] + '%|\n')
        string.append('|Unexecuted lines            |' + gcov.metrics['coverage'][2] + '|' + gcov.metrics['coverage'][4] + '%|\n')
        string.append('|Average hits / executed     |' + gcov.metrics['coverage'][5] + '| |\n')
        string.append('\n')
        string.append(_mermaid_pie('Lines (' + gcov.metrics['coverage'][3] + '% covered)',
                                   gcov.metrics['coverage'][1], gcov.metrics['coverage'][2]))
      if gcov.metrics['procedures']:
        string.append('|Procedures| | |\n')
        string.append('| --- | --- | --- |\n')
        string.append('|Total procedures            |' + gcov.metrics['procedures'][0] + '| |\n')
        string.append('|Executed procedures         |' + gcov.metrics['procedures'][1] + '|' + gcov.metrics['procedures'][3] + '%|\n')
        string.append('|Unexecuted procedures       |' + gcov.metrics['procedures'][2] + '|' + gcov.metrics['procedures'][4] + '%|\n')
        string.append('|Average hits / executed     |' + gcov.metrics['procedures'][5] + '| |\n')
        string.append('\n')
        string.append(_mermaid_pie('Procedures (' + gcov.metrics['procedures'][3] + '% covered)',
                                   gcov.metrics['procedures'][1], gcov.metrics['procedures'][2]))
    with open(os.path.join(configuration.cliargs.gcov_analyzer[0], configuration.cliargs.gcov_analyzer[1] + '.md'), 'w') as summary:
      summary.writelines(string)


def is_ascii_kind_supported(configuration):
  """Check is the compiler support ASCII character kind.

  Parameters
  ----------
  configuration : FoBiSConfig()


  Returns
  -------
  bool
    true if ASCII kind is supported, false otherwise
  """
  builder = Builder(cliargs=configuration.cliargs, print_n=configuration.print_b, print_w=configuration.print_r)
  test_file_name = os.path.join(builder.build_dir, 'ascii_kind_introspection.f90')
  test = open(test_file_name, 'w')
  test.write("program test\nprint*, selected_char_kind('ascii')\nendprogram")
  test.close()
  test = ParsedFile(name=test_file_name)
  test.parse(inc=configuration.cliargs.inc)
  pfiles = [test]
  dependency_hiearchy(builder=builder, pfiles=pfiles, print_w=configuration.print_r)
  builder.build(file_to_build=pfiles[0], verbose=configuration.cliargs.verbose, log=configuration.cliargs.log)
  os.remove(test_file_name)
  test_exe = os.path.join(builder.build_dir, os.path.splitext(os.path.basename(test.name))[0])
  result = syswork(test_exe)
  os.remove(test_exe)
  is_supported = False
  if result[0] == 0:
    if int(result[1]) > 0:
      is_supported = True
  print("Compiler '" + builder.compiler.compiler + "' support ASCII kind:", is_supported)
  return is_supported


def is_ucs4_kind_supported(configuration):
  """Check is the compiler support UCS4 character kind.

  Parameters
  ----------
  configuration : FoBiSConfig()


  Returns
  -------
  bool
    true if UCS4 kind is supported, false otherwise
  """
  builder = Builder(cliargs=configuration.cliargs, print_n=configuration.print_b, print_w=configuration.print_r)
  test_file_name = os.path.join(builder.build_dir, 'ucs4_kind_introspection.f90')
  test = open(test_file_name, 'w')
  test.write("program test\nprint*, selected_char_kind('iso_10646')\nendprogram")
  test.close()
  test = ParsedFile(name=test_file_name)
  test.parse(inc=configuration.cliargs.inc)
  pfiles = [test]
  dependency_hiearchy(builder=builder, pfiles=pfiles, print_w=configuration.print_r)
  builder.build(file_to_build=pfiles[0], verbose=configuration.cliargs.verbose, log=configuration.cliargs.log)
  os.remove(test_file_name)
  test_exe = os.path.join(builder.build_dir, os.path.splitext(os.path.basename(test.name))[0])
  result = syswork(test_exe)
  os.remove(test_exe)
  is_supported = False
  if result[0] == 0:
    if int(result[1]) > 0:
      is_supported = True
  print("Compiler '" + builder.compiler.compiler + "' support UCS4 kind:", is_supported)
  return is_supported


def is_float128_kind_supported(configuration):
  """Check is the compiler support float128 real kind.

  Parameters
  ----------
  configuration : FoBiSConfig()


  Returns
  -------
  bool
    true if UCS4 kind is supported, false otherwise
  """
  builder = Builder(cliargs=configuration.cliargs, print_n=configuration.print_b, print_w=configuration.print_r)
  test_file_name = os.path.join(builder.build_dir, 'float128_kind_introspection.f90')
  test = open(test_file_name, 'w')
  test.write("program test\nprint*, selected_real_kind(33,4931)\nendprogram")
  test.close()
  test = ParsedFile(name=test_file_name)
  test.parse(inc=configuration.cliargs.inc)
  pfiles = [test]
  dependency_hiearchy(builder=builder, pfiles=pfiles, print_w=configuration.print_r)
  builder.build(file_to_build=pfiles[0], verbose=configuration.cliargs.verbose, log=configuration.cliargs.log)
  os.remove(test_file_name)
  test_exe = os.path.join(builder.build_dir, os.path.splitext(os.path.basename(test.name))[0])
  result = syswork(test_exe)
  os.remove(test_exe)
  is_supported = False
  if result[0] == 0:
    if int(result[1]) > 0:
      is_supported = True
  print("Compiler '" + builder.compiler.compiler + "' support float128 kind:", is_supported)
  return is_supported

if __name__ == '__main__':
  main()



#!/usr/bin/env python
"""Setup script for install FoBiS.py"""
# import pypandoc
import re
from setuptools import setup
__source__ = open('fobis/FoBiSConfig.py').read()
__license__ = re.search(r'^__license__\s*=\s*"(.*)"', __source__, re.M).group(1)

if __name__ == '__main__':
  setup(name                  = re.search(r'^__appname__\s*=\s*"(.*)"', __source__, re.M).group(1),
        version               = re.search(r'^__version__\s*=\s*"(.*)"', __source__, re.M).group(1),
        description           = re.search(r'^__description__\s*=\s*"(.*)"', __source__, re.M).group(1),
        long_description      = re.search(r'^__long_description__\s*=\s*"(.*)"', __source__, re.M).group(1),
        author                = re.search(r'^__author__\s*=\s*"(.*)"', __source__, re.M).group(1),
        author_email          = re.search(r'^__author_email__\s*=\s*"(.*)"', __source__, re.M).group(1),
        url                   = re.search(r'^__url__\s*=\s*"(.*)"', __source__, re.M).group(1),
        scripts               = ['FoBiS.py'],
        packages              = ['fobis'],
        py_modules            = [],
        classifiers           = ['Development Status :: 5 - Production/Stable',
                                 'License :: OSI Approved :: '+__license__,
                                 'Environment :: Console',
                                 'Intended Audience :: End Users/Desktop',
                                 'Programming Language :: Python',
                                 'Programming Language :: Python :: 2',
                                 'Programming Language :: Python :: 2.7',
                                 'Programming Language :: Python :: 3',
                                 'Programming Language :: Python :: 3.8',
                                 'Topic :: Text Processing'],
        entry_points          = { 'console_scripts': ['fobis = fobis.fobis:main'] },
        extras_require        = { 'PreForM.py':  ["PreForM.py>=v1.1.1"],
                                  'FORD': ["FORD>=1.1.0"],
                                  'graphviz': ["graphviz>=0.4.2"]},
        install_requires      = [ "future", "configparser", "typer[all]>=0.9.0", "colorama ; platform_system=='Windows'" ])



"""
Colors.py, module definition of Colors class.
This is a class aimed at coloring prints.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from __future__ import print_function
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import unicode_literals


# from future import standard_library
# standard_library.install_aliases()
# from builtins import *
# from builtins import object
class Colors(object):
  """
  Colors is an object that handles colors of shell prints, its attributes and methods.
  """
  def __init__(self,
               red='\033[1;31m',
               bld='\033[1m',
               enabled=True):
    self.red = red
    self.bld = bld
    self.end = '\033[0m'
    if not enabled:
      self.disable()

  def enable(self):
    """Method for enabling colors."""
    self.red = '\033[1;31m'
    self.bld = '\033[1m'
    self.end = '\033[0m'

  def disable(self):
    """Method for disabling colors."""
    self.red = ''
    self.bld = ''
    self.end = ''

  def print_b(self, string, end="\n"):
    """
    Method for printing string with bold color.

    Parameters
    ----------
    string : str
      string to be printed
    """
    print(self.bld + string + self.end, end=end)
    return

  def print_r(self, string, end='\n'):
    """
    Method for printing string with red color.

    Parameters
    ----------
    string : str
      string to be printed
    """
    print(self.red + string + self.end, end=end)
    return
